import pytest
from flexmock import flexmock

import packit
from packit.copr_helper import CoprHelper


class TestCoprHelper:
    @pytest.mark.parametrize(
        # copr_client.mock_chroot_proxy.get_list() returns dictionary
        "get_list_keys, expected_return",
        [
            pytest.param(["chroot1", "_chroot2"], ["chroot1"], id="chroot_list"),
            pytest.param([], [], id="empty_list"),
        ],
    )
    def test_get_avilable_chroots(self, get_list_keys, expected_return):
        copr_client_mock = flexmock(mock_chroot_proxy=flexmock())
        copr_client_mock.mock_chroot_proxy.should_receive("get_list.keys").and_return(
            get_list_keys
        )
        flexmock(packit.copr_helper.CoprClient).should_receive(
            "create_from_config_file"
        ).and_return(copr_client_mock)

        copr_helper = CoprHelper("_upstream_local_project")
        copr_helper.get_available_chroots.cache_clear()

        assert copr_helper.get_available_chroots() == expected_return

    @pytest.mark.parametrize(
        "owner,project,section,expected_suffix",
        [
            (
                "@rhinstaller",
                "Anaconda",
                "permissions",
                "g/rhinstaller/Anaconda/permissions/",
            ),
            ("@rhinstaller", "Anaconda", None, "g/rhinstaller/Anaconda/edit/"),
            ("someone", "Anaconda", "permissions", "someone/Anaconda/permissions/"),
        ],
    )
    def test_settings_url(self, owner, project, section, expected_suffix):
        copr_client_mock = flexmock(config={"copr_url": "https://fedoracloud.org"})

        flexmock(packit.copr_helper.CoprClient).should_receive(
            "create_from_config_file"
        ).and_return(copr_client_mock)
        copr_helper = CoprHelper("_upstream_local_project")

        assert (
            copr_helper.get_copr_settings_url(owner, project, section)
            == f"https://fedoracloud.org/coprs/{expected_suffix}"
        )



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import inspect
import re
from logging import getLogger
from pathlib import Path
from typing import Union, List, Optional, Dict

from packit.patches import PatchMetadata
from rebasehelper.helpers.macro_helper import MacroHelper
from rebasehelper.specfile import SpecFile, RebaseHelperError, saves, PatchObject
from rebasehelper.tags import Tag, Tags

try:
    from rebasehelper.plugins.plugin_manager import plugin_manager
except ImportError:
    from rebasehelper.versioneer import versioneers_runner

from packit.exceptions import PackitException

logger = getLogger(__name__)


class Specfile(SpecFile):
    def __init__(self, path: Union[str, Path], sources_dir: Union[str, Path] = ""):
        s = inspect.signature(SpecFile)
        if "changelog_entry" in s.parameters:
            super().__init__(
                path=str(path), sources_location=str(sources_dir), changelog_entry=""
            )
        else:
            super().__init__(path=str(path), sources_location=str(sources_dir))

    def update_spec(self):
        if hasattr(self, "update"):
            # new rebase-helper
            self.update()
        else:
            # old rebase-helper
            self._update_data()

    def update_changelog_in_spec(self, changelog_entry):
        if hasattr(self, "update_changelog"):
            # new rebase-helper
            self.update_changelog(changelog_entry)
        else:
            # old rebase-helper
            self.changelog_entry = changelog_entry
            new_log = self.get_new_log()
            new_log.extend(self.spec_content.sections["%changelog"])
            self.spec_content.sections["%changelog"] = new_log
            self.save()

    def set_spec_version(
        self, version: str = None, release: str = None, changelog_entry: str = None
    ):
        """
        Set version in spec, release and add a changelog_entry (if they are presented).

        :param version: new version
        :param release: new release
        :param changelog_entry: accompanying changelog entry
        """
        try:
            if version:
                # also this code adds 3 rpmbuild dirs into the upstream repo,
                # we should ask rebase-helper not to do that
                # using set_tag instead of set_version to turn off preserving macros
                self.set_tag("Version", version, preserve_macros=False)

            if release:
                # using set_tag instead of set_release to turn off preserving macros
                self.set_tag(
                    "Release", "{}%{{?dist}}".format(release), preserve_macros=False
                )

            if not changelog_entry:
                return

            if not self.spec_content.section("%changelog"):
                logger.debug(
                    "The specfile doesn't have any %changelog, will not set it."
                )
                return

            self.update_changelog_in_spec(changelog_entry)

        except RebaseHelperError as ex:
            logger.error(f"Rebase-helper failed to change the spec file: {ex}")
            raise PackitException("Rebase-helper didn't do the job.")

    def write_spec_content(self):
        if hasattr(self, "_write_spec_content"):
            # new rebase-helper
            self._write_spec_content()
        else:
            # old rebase-helper
            self._write_spec_file_to_disc()

    @staticmethod
    def get_upstream_version(versioneer, package_name, category):
        """
        Call the method of rebase-helper (due to the version of rebase-helper)
        to get the latest upstream version of a package.
        :param versioneer:
        :param package_name: str
        :param category:
        :return: str version
        """
        try:
            get_version = plugin_manager.versioneers.run
        except NameError:
            get_version = versioneers_runner.run
        return get_version(versioneer, package_name, category)

    def get_release_number(self) -> str:
        """
        Removed in rebasehelper=0.20.0
        """
        release = self.header.release
        dist = MacroHelper.expand("%{dist}")
        if dist:
            release = release.replace(dist, "")
        return re.sub(r"([0-9.]*[0-9]+).*", r"\1", release)

    @saves
    def set_patches(self, patch_list: List[PatchMetadata]) -> None:
        """
        Set given patches in the spec file

        :param patch_list: [PatchMetadata]
        """
        if not patch_list:
            return

        if all(p.present_in_specfile for p in patch_list):
            logger.debug(
                "All patches are present in the spec file, nothing to do here 🚀"
            )
            return

        # we could have generated patches before (via git-format-patch)
        # so let's reload the spec
        self.reload()

        applied_patches: Dict[str, PatchObject] = {
            p.get_patch_name(): p for p in self.get_applied_patches()
        }

        for patch_metadata in patch_list:
            if patch_metadata.present_in_specfile:
                logger.debug(
                    f"Patch {patch_metadata.name} is already present in the spec file."
                )
                continue

            if patch_metadata.name in applied_patches:
                logger.debug(
                    f"Patch {patch_metadata.name} is already defined in the spec file."
                )
                continue

            self.add_patch(patch_metadata)

    def add_patch(self, patch_metadata: PatchMetadata):
        """
        Add provided patch to the spec file:
         * Set Patch index to be +1 than the highest index of an existing specfile patch
         * The Patch placement logic works like this:
           * If there already are patches, then the patch is added after them
           * If there are no existing patches, the patch is added after Source definitions
        """
        try:
            patch_number_offset = max(x.index for x in self.get_applied_patches())
        except ValueError:
            logger.debug("There are no patches in the spec.")
            patch_number_offset = 0

        new_content = "\n# " + "\n# ".join(patch_metadata.specfile_comment.split("\n"))
        new_content += f"\nPatch{(1 + patch_number_offset):04d}: {patch_metadata.name}"

        if self.get_applied_patches():
            last_source_tag_line = [
                t.line for t in self.tags.filter(name="Patch*", valid=None)
            ][-1]
        else:
            last_source_tag_line = [
                t.line for t in self.tags.filter(name="Source*", valid=None)
            ][-1]

        # Find first empty line after last_source_tag_line
        for index, line in enumerate(
            self.spec_content.section("%package")[last_source_tag_line:],
            start=last_source_tag_line,
        ):
            if not line:
                where = index
                break
        else:
            where = len(self.spec_content.section("%package"))

        logger.debug(f"Adding patch {patch_metadata.name} to the spec file.")
        self.spec_content.section("%package")[where:where] = new_content.split("\n")
        self.save()

    def get_source(self, source_name: str) -> Optional[Tag]:
        """
        get specific Source from spec

        :param source_name: precise name of the Source, e.g. Source1, or Source
        :return: corresponding Source Tag
        """
        # sanitize the name, this will also add index if there isn't one
        source_name, *_ = Tags._sanitize_tag(source_name, 0, 0)
        return next(self.tags.filter(name=source_name, valid=None), None)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import fileinput
import subprocess
from pathlib import Path
import yaml

import pytest

from packit.constants import CENTOS_DOMAIN, CENTOS_STREAM_GITLAB
from packit.fedpkg import FedPKG
from packit.local_project import LocalProject
from packit.patches import PatchMetadata
from packit.source_git import SourceGitGenerator
from packit.specfile import Specfile
from packit.utils.repo import create_new_repo, clone_centos_9_package
from tests.spellbook import initiate_git_repo

UNIVERSAL_PACKAGE_NAME = "redhat-rpm-config"


@pytest.mark.parametrize(
    "fedora_package, centos_package, branch",
    (
        (UNIVERSAL_PACKAGE_NAME, None, "main"),
        (None, UNIVERSAL_PACKAGE_NAME, "c8s"),
    ),
)
def test_distgit_cloning(
    api_instance_source_git, fedora_package, centos_package, branch, tmp_path: Path
):
    sgg = SourceGitGenerator(
        api_instance_source_git.upstream_local_project,
        api_instance_source_git.config,
        # yes, this is the upstream repo
        f"https://src.fedoraproject.org/rpms/{UNIVERSAL_PACKAGE_NAME}",
        fedora_package=fedora_package,
        centos_package=centos_package,
        dist_git_branch=branch,
        tmpdir=tmp_path,
    )
    sgg._get_dist_git()
    assert tmp_path.joinpath(UNIVERSAL_PACKAGE_NAME, ".git").is_dir()


# TODO: test different refs: tag, branch, commit hash
def test_fetch_upstream_ref(api_instance_source_git, tmp_path: Path):
    tag = "1.0.0"
    s = tmp_path.joinpath("s")
    u = tmp_path.joinpath("u")
    u.mkdir()
    create_new_repo(Path(s), [])
    initiate_git_repo(u, tag=tag)
    sgg = SourceGitGenerator(
        LocalProject(working_dir=s),
        api_instance_source_git.config,
        str(u),
        upstream_ref=tag,
        centos_package="x",
    )

    sgg._pull_upstream_ref()

    assert s.joinpath(".git").is_dir()
    assert sgg.local_project.ref == "main"
    assert sgg.local_project.working_dir.joinpath("hops").read_text() == "Cascade\n"
    assert sgg.local_project.git_repo.head.commit.message == "commit with data\n"


@pytest.mark.parametrize(
    "fedora_package, centos_package, branch",
    (("fuse-overlayfs", None, "main"),),
)
def test_run_prep(
    api_instance_source_git, fedora_package, centos_package, branch, tmp_path: Path
):
    sgg = SourceGitGenerator(
        api_instance_source_git.upstream_local_project,
        api_instance_source_git.config,
        f"https://github.com/containers/{fedora_package}",
        fedora_package=fedora_package,
        centos_package=centos_package,
        dist_git_branch=branch,
        tmpdir=tmp_path,
    )
    assert sgg.primary_archive.exists()  # making sure this is downloaded
    sgg._run_prep()
    build_dir = sgg.dist_git.local_project.working_dir.joinpath("BUILD")
    assert build_dir.exists()
    project_dir = next(build_dir.glob("fuse-overlayfs-*"))
    assert project_dir.joinpath(".git")


def test_create_packit_yaml_upstream_project_url(
    api_instance_source_git, tmp_path: Path
):
    """
    use requre to create a source-git out of it in an empty git repo - packit
    will pull upstream git history
    """
    # requre upstream_project_url
    upstream_project_url = "https://github.com/packit/requre.git"

    # clone dist-git
    pkg = "python-requre"
    dist_git_ref = "6b27ffacda06289ca2d546e15b3c96845243005f"
    dist_git_path = tmp_path.joinpath(pkg)
    source_git_path = tmp_path.joinpath("requre-sg")
    FedPKG().clone(pkg, str(dist_git_path), anonymous=True)

    # check out specific ref
    subprocess.check_call(["git", "reset", "--hard", dist_git_ref], cwd=dist_git_path)

    # create src-git
    source_git_path.mkdir()
    create_new_repo(Path(source_git_path), [])
    sgg = SourceGitGenerator(
        LocalProject(working_dir=source_git_path),
        api_instance_source_git.config,
        upstream_project_url,
        upstream_ref="0.4.0",
        dist_git_path=dist_git_path,
    )
    sgg.create_from_upstream()

    config_file = Path(source_git_path / ".packit.yaml").read_text()
    # black sucks here :/ we are making sure here the yaml looks nice
    assert "patch_generation_ignore_paths:\n- .distro\n" in config_file
    assert "sources:\n- path: requre-0.4.0.tar.gz\n" in config_file
    packit_yaml = yaml.load(config_file)
    assert packit_yaml.get("upstream_project_url") == upstream_project_url


def test_create_packit_yaml_sources(api_instance_source_git, tmp_path: Path):
    """
    use requre to create a source-git out of it in an empty git repo - packit
    will pull upstream git history
    """
    # requre lookaside_cache_url
    requre_tar_url = (
        "https://src.fedoraproject.org/repo/pkgs/python-requre/"
        "requre-0.4.0.tar.gz/sha512/85293577f56e19dd0fad13bb5e118ac2ab39d7"
        "570d640a754b9d8d8054078c89c949d4695ef018915b17a2f2428f1635032352d"
        "cf3c9a036a2d633013cc35dd9/requre-0.4.0.tar.gz"
    )
    requre_tar_path = "requre-0.4.0.tar.gz"

    # clone dist-git
    pkg = "python-requre"
    dist_git_ref = "6b27ffacda06289ca2d546e15b3c96845243005f"
    dist_git_path = tmp_path.joinpath(pkg)
    source_git_path = tmp_path.joinpath("requre-sg")
    FedPKG().clone(pkg, str(dist_git_path), anonymous=True)

    # check out specific ref
    subprocess.check_call(["git", "reset", "--hard", dist_git_ref], cwd=dist_git_path)

    # create src-git
    source_git_path.mkdir()
    create_new_repo(Path(source_git_path), [])
    sgg = SourceGitGenerator(
        LocalProject(working_dir=source_git_path),
        api_instance_source_git.config,
        "https://github.com/packit/requre",
        upstream_ref="0.4.0",
        dist_git_path=dist_git_path,
    )
    sgg.create_from_upstream()

    config_file = Path(source_git_path / ".packit.yaml").read_text()
    # black sucks here :/ we are making sure here the yaml looks nice
    assert "patch_generation_ignore_paths:\n- .distro\n" in config_file
    assert "sources:\n- path: requre-0.4.0.tar.gz\n" in config_file
    packit_yaml = yaml.load(config_file)
    assert packit_yaml.get("sources")
    assert len(packit_yaml["sources"]) > 0
    assert packit_yaml["sources"][0].get("url")
    assert packit_yaml["sources"][0].get("path")

    assert packit_yaml["sources"][0]["url"] == requre_tar_url
    assert packit_yaml["sources"][0]["path"] == requre_tar_path


REQURE_PATCH = r"""\
diff --git a/README.md b/README.md
index 17e9e85..99ef68c 100644
--- a/README.md
+++ b/README.md
@@ -15,3 +15,5 @@ back
 - Used for testing [packit-service](https://github.com/packit-service) organization projects
   - ogr
   - packit
+
+Hello!
"""


def test_create_srcgit_requre_clean(api_instance_source_git, tmp_path: Path):
    """
    use requre to create a source-git out of it in an empty git repo - packit
    will pull upstream git history
    """
    # clone dist-git
    pkg = "python-requre"
    dist_git_ref = "6b27ffacda06289ca2d546e15b3c96845243005f"
    dist_git_path = tmp_path.joinpath(pkg)
    source_git_path = tmp_path.joinpath("requre-sg")
    FedPKG().clone(pkg, str(dist_git_path), anonymous=True)
    dg_lp = LocalProject(working_dir=dist_git_path)

    # check out specific ref
    subprocess.check_call(["git", "reset", "--hard", dist_git_ref], cwd=dist_git_path)

    # add a patch in there
    spec = Specfile(dist_git_path / f"{pkg}.spec", sources_dir=dist_git_path)
    patch_name = "hello.patch"
    patch_path = dist_git_path.joinpath(patch_name)
    patch_path.write_text(REQURE_PATCH)
    patch = PatchMetadata(name=patch_name, path=patch_path, present_in_specfile=False)
    spec.add_patch(patch)
    dg_lp.stage()
    dg_lp.commit("add the hello patch")
    subprocess.check_call(["fedpkg", "prep"], cwd=dist_git_path)

    # create src-git
    source_git_path.mkdir()
    create_new_repo(Path(source_git_path), [])
    sgg = SourceGitGenerator(
        LocalProject(working_dir=source_git_path),
        api_instance_source_git.config,
        "https://github.com/packit/requre",
        upstream_ref="0.4.0",
        dist_git_path=dist_git_path,
    )
    sgg.create_from_upstream()

    # verify it
    subprocess.check_call(["packit", "srpm"], cwd=source_git_path)
    srpm_path = list(source_git_path.glob("python-requre-0.4.0-2.*.src.rpm"))[0]
    assert srpm_path.is_file()
    # requre needs sphinx, so SRPM is fine

    # verify the archive is not committed in the source-git
    with pytest.raises(subprocess.CalledProcessError) as exc:
        subprocess.check_call(
            [
                "git",
                "ls-files",
                "--error-unmatch",
                f"{sgg.dist_git.source_git_downstream_suffix}/{spec.get_archive()}",
            ],
            cwd=source_git_path,
        )
    assert exc.value.returncode == 1


def test_create_srcgit_requre_populated(api_instance_source_git, tmp_path: Path):
    """
    use requre to create a source-git out of it in a branch with upstream git history
    - this should only layer downstream changes on top
    """
    # clone dist-git
    pkg = "python-requre"
    dist_git_ref = "6b27ffacda06289ca2d546e15b3c96845243005f"
    dist_git_path = tmp_path.joinpath(pkg)
    source_git_path = tmp_path.joinpath("requre-sg")
    FedPKG().clone(pkg, str(dist_git_path), anonymous=True)
    dg_lp = LocalProject(working_dir=dist_git_path)

    # check out specific ref
    subprocess.check_call(["git", "reset", "--hard", dist_git_ref], cwd=dist_git_path)

    # add a patch in there
    spec = Specfile(dist_git_path / f"{pkg}.spec", sources_dir=dist_git_path)
    patch_name = "hello.patch"
    patch_path = dist_git_path.joinpath(patch_name)
    patch_path.write_text(REQURE_PATCH)
    patch = PatchMetadata(name=patch_name, path=patch_path, present_in_specfile=False)
    spec.add_patch(patch)
    dg_lp.stage()
    dg_lp.commit("add the hello patch")
    subprocess.check_call(["fedpkg", "prep"], cwd=dist_git_path)

    # create src-git
    source_git_path.mkdir()
    subprocess.check_call(
        ["git", "clone", "https://github.com/packit/requre", str(source_git_path)]
    )
    subprocess.check_call(
        ["git", "checkout", "-B", "source-git-0.4.0", "0.4.0"], cwd=source_git_path
    )
    sgg = SourceGitGenerator(
        LocalProject(working_dir=source_git_path),
        api_instance_source_git.config,
        dist_git_path=dist_git_path,
    )
    sgg.create_from_upstream()

    # verify it
    subprocess.check_call(["packit", "srpm"], cwd=source_git_path)
    srpm_path = list(source_git_path.glob("python-requre-0.4.0-2.*.src.rpm"))[0]
    assert srpm_path.is_file()
    # requre needs sphinx, so SRPM is fine


@pytest.mark.slow
@pytest.mark.parametrize(
    "dist_git_branch,upstream_ref", (("c8s", "cronie-1.5.2"), ("c9s", "cronie-1.5.5"))
)
def test_centos_cronie(
    dist_git_branch, upstream_ref, api_instance_source_git, tmp_path: Path
):
    source_git_path = tmp_path.joinpath("cronie-sg")
    # create src-git
    source_git_path.mkdir()
    create_new_repo(source_git_path, [])
    sgg = SourceGitGenerator(
        LocalProject(working_dir=source_git_path),
        api_instance_source_git.config,
        "https://github.com/cronie-crond/cronie",
        upstream_ref=upstream_ref,
        centos_package="cronie",
        dist_git_branch=dist_git_branch,
    )
    sgg.create_from_upstream()

    if dist_git_branch == "c8s":
        assert CENTOS_DOMAIN in sgg.dist_git.local_project.git_url
    else:
        assert CENTOS_STREAM_GITLAB in sgg.dist_git.local_project.git_url

    # verify it
    subprocess.check_call(["packit", "srpm"], cwd=source_git_path)
    srpm_path = list(source_git_path.glob("cronie-*.src.rpm"))[0]
    assert srpm_path.is_file()


@pytest.mark.slow
@pytest.mark.parametrize("apply_option", ("git", "git_am"))
def test_acl_with_git_git_am(apply_option, api_instance_source_git, tmp_path: Path):
    """manipulate acl's dist-git to use -Sgit and -Sgit_am so we can verify it works"""
    dist_git_branch = "c9s"
    package_name = "acl"
    source_git_path = tmp_path.joinpath("acl-sg")
    # dist-git tools expect: dir name == package name
    dist_git_path = tmp_path.joinpath(package_name)

    # create src-git
    source_git_path.mkdir()
    create_new_repo(source_git_path, [])

    # fetch dist-git and change %autosetup
    clone_centos_9_package(
        package_name,
        dist_git_path=dist_git_path,
        branch=dist_git_branch,
    )
    for line in fileinput.input(
        dist_git_path.joinpath(f"{package_name}.spec"), inplace=True
    ):
        if "%autosetup" in line:
            line = f"%autosetup -p1 -S{apply_option}"
        print(line, end="")  # \n would make double-newlines here

    # run conversion
    sgg = SourceGitGenerator(
        LocalProject(working_dir=source_git_path),
        api_instance_source_git.config,
        "https://git.savannah.nongnu.org/git/acl.git",
        upstream_ref="v2.3.1",
        centos_package=package_name,
        dist_git_branch=dist_git_branch,
        dist_git_path=dist_git_path,
    )
    sgg.create_from_upstream()

    # verify the patch commit has metadata
    patch_commit_message = sgg.local_project.git_repo.head.commit.message
    assert "present_in_specfile: true" in patch_commit_message
    assert "patch_name: 0001-acl-2.2.53-test-runwrapper.patch" in patch_commit_message



# MIT License
#
# Copyright (c) 2020 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging

import pytest
from packit.config.config import Config

logger = logging.getLogger(__name__)

dep_keys = [
    "github_app_id",
    "github_app_cert_path",
    "github_token",
    "pagure_user_token",
    "pagure_instance_url",
    "pagure_fork_token",
]

authentication = {
    "authentication": {
        "github.com": {"token": "abcd"},
        "pagure": {"token": "abcd", "instance_url": "https://src.fedoraproject.org"},
    }
}


@pytest.mark.parametrize(
    "key",
    dep_keys,
)
def test_with_deprecated_keys(key, caplog, recwarn):
    raw_dict = {key: "somevalue"}
    Config.load_authentication(raw_dict)
    assert "Please, use 'authentication' key in the user" in caplog.text
    caplog.clear()


def test_with_only_authentication(caplog, recwarn):
    Config.load_authentication(authentication)
    assert "Please, use 'authentication' key in the user" not in caplog.text


def test_with_both_authentication(caplog, recwarn):
    raw_dict = authentication
    raw_dict[dep_keys[3]] = "somevalue"
    Config.load_authentication(raw_dict)
    assert "Please, use 'authentication' key in the user" not in caplog.text



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from packit.utils.commands import cwd

from tests.spellbook import call_packit


def test_generate_pass(upstream_without_config):
    with cwd(upstream_without_config):
        assert not (upstream_without_config / ".packit.yaml").is_file()

        # This test requires packit on pythonpath
        result = call_packit(parameters=["generate"])

        assert result.exit_code == 0

        assert (upstream_without_config / ".packit.yaml").is_file()


def test_generate_fail(cwd_upstream_or_distgit):
    result = call_packit(
        parameters=["generate"], working_dir=str(cwd_upstream_or_distgit)
    )

    assert result.exit_code == 2  # packit config already exists --force needed



# MIT License
#
# Copyright (c) 2018 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
from flexmock import flexmock
from gnupg import GPG

from packit.api import PackitAPI
from packit.exceptions import PackitException
from packit.security import CommitVerifier
from tests.integration.utils import remove_gpg_key_pair


def test_allowed_gpg_keys_none(api_instance_source_git: PackitAPI):
    api_instance_source_git.up.allowed_gpg_keys = None
    flexmock(CommitVerifier).should_receive("check_signature_of_commit").times(0)
    api_instance_source_git.up.check_last_commit()


@pytest.mark.parametrize("allowed_gpg_keys", [[], ["abcd", "efgh"]])
def test_allowed_gpg_keys_not_allowed(
    allowed_gpg_keys, api_instance_source_git: PackitAPI, gnupg_key_fingerprint: str
):
    api_instance_source_git.up.local_project.git_repo.git.commit(
        message="signed commit", gpg_sign=gnupg_key_fingerprint, allow_empty=True
    )

    api_instance_source_git.up.allowed_gpg_keys = allowed_gpg_keys
    with pytest.raises(PackitException) as ex:
        api_instance_source_git.up.check_last_commit()
    assert "not signed" in str(ex)


def test_allowed_gpg_keys_allowed(
    api_instance_source_git: PackitAPI, gnupg_key_fingerprint: str
):
    api_instance_source_git.up.local_project.git_repo.git.commit(
        message="signed commit", gpg_sign=gnupg_key_fingerprint, allow_empty=True
    )

    api_instance_source_git.up.allowed_gpg_keys = [gnupg_key_fingerprint]
    api_instance_source_git.up.check_last_commit()


def test_allowed_gpg_keys_not_existing_key(
    api_instance_source_git: PackitAPI, gnupg_instance: GPG, gnupg_key_fingerprint: str
):
    api_instance_source_git.up.local_project.git_repo.git.commit(
        message="signed commit", gpg_sign=gnupg_key_fingerprint, allow_empty=True
    )
    remove_gpg_key_pair(
        gpg_binary=gnupg_instance.gpgbinary, fingerprint=gnupg_key_fingerprint
    )
    api_instance_source_git.up.allowed_gpg_keys = [gnupg_key_fingerprint]
    with pytest.raises(PackitException) as ex:
        api_instance_source_git.up.check_last_commit()
    assert "Cannot receive a gpg key" in str(ex)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
from pathlib import Path

from flexmock import flexmock

from packit.config import Config
from packit.distgit import DistGit
from packit.local_project import LocalProject
from packit.source_git import SourceGitGenerator, CentOS8DistGit
from packit.utils.repo import create_new_repo


def test_centos_conf(cronie, tmp_path: Path):
    """make sure the centos-specific configuration is correct"""
    source_git_path = tmp_path.joinpath("cronie-sg")
    # create src-git
    source_git_path.mkdir()
    create_new_repo(source_git_path, [])
    sgg = SourceGitGenerator(
        LocalProject(working_dir=source_git_path),
        Config(),
        dist_git_path=cronie,
        upstream_ref="cronie-1.5.2",
        centos_package="cronie",
    )

    dg = sgg.dist_git
    assert isinstance(dg, CentOS8DistGit)

    flexmock(
        DistGit,
        download_upstream_archive=lambda: cronie / "SOURCES" / "cronie-1.5.2.tar.gz",
    )
    assert sgg.primary_archive == cronie / "SOURCES" / "cronie-1.5.2.tar.gz"

    assert dg.absolute_source_dir == cronie / "SOURCES"
    assert dg.absolute_specfile_dir == cronie / "SPECS"
    assert dg.absolute_specfile_path == cronie / "SPECS" / "cronie.spec"



# MIT License
#
# Copyright (c) 2020 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Functional tests for local-build command
"""
from pathlib import Path
from subprocess import CalledProcessError

import pytest
from distro import linux_distribution

from tests.functional.spellbook import call_real_packit


def test_rpm_command(ogr_distgit_and_remote):
    call_real_packit(
        parameters=["--debug", "local-build"], cwd=ogr_distgit_and_remote[0]
    )
    rpm_paths = ogr_distgit_and_remote[0].glob("noarch/*.rpm")

    assert all(rpm_path.exists() for rpm_path in rpm_paths)


def test_local_build_with_remote_good(ogr_distgit_and_remote):
    call_real_packit(
        parameters=["--debug", "--remote", "origin", "local-build"],
        cwd=ogr_distgit_and_remote[0],
    )
    rpm_paths = ogr_distgit_and_remote[0].glob("noarch/*.rpm")

    assert all(rpm_path.exists() for rpm_path in rpm_paths)


def test_local_build_with_remote_bad(ogr_distgit_and_remote):
    with pytest.raises(CalledProcessError) as ex:
        call_real_packit(
            parameters=["--debug", "--remote", "burcak", "local-build"],
            cwd=ogr_distgit_and_remote[0],
            return_output=True,
        )
    assert b"Remote named 'burcak' didn't exist" in ex.value.output


@pytest.mark.skipif(
    linux_distribution()[0].startswith("CentOS"),
    reason="No rpmautospec-rpm-macros installed",
)
def test_rpm_command_for_path(ogr_distgit_and_remote):
    call_real_packit(
        parameters=["--debug", "local-build", str(ogr_distgit_and_remote[0])]
    )
    rpm_paths = Path.cwd().glob("noarch/*.rpm")
    assert all(rpm_path.exists() for rpm_path in rpm_paths)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Functions and classes dealing with syncing files between repositories
"""

import glob
import logging
from pathlib import Path
from typing import List, Optional, Union, Sequence

from packit.exceptions import PackitException
from packit.utils import run_command

logger = logging.getLogger(__name__)


def check_subpath(subpath: Path, path: Path) -> Path:
    """Check if 'subpath' is a subpath of 'path'

    Args:
        subpath: Subpath to be checked.
        path: Path agains which subpath is checked.

    Returns:
        'subpath', resolved, in case it is a subpath of 'path'.

    Raises:
        PackitException, if 'subpath' is not a subpath of 'path'.
    """
    if not subpath.resolve().is_relative_to(path.resolve()):
        raise PackitException(
            f"Sync files: Illegal path! {subpath} is not in the subpath of {path}."
        )
    return subpath.resolve()


class SyncFilesItem:
    """Some files to sync to destination

    Think about this as a wrapper around 'rsync'.

    Attributes:
        src: List of paths to sync.
        dest: Destination to sync to.
        mkpath: Create the destination's path component.
    """

    def __init__(
        self,
        src: Sequence[Union[str, Path]],
        dest: Union[str, Path],
        mkpath: bool = False,
    ):
        self.src = [Path(s) for s in src]
        self.dest = Path(dest)
        self.mkpath = mkpath

    def __repr__(self):
        return f"SyncFilesItem(src={self.src}, dest={self.dest}, mkpath={self.mkpath})"

    def __str__(self):
        return " ".join(self.command())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SyncFilesItem):
            raise NotImplementedError()

        return self.src == other.src and self.dest == other.dest

    def command(self, fail_on_missing: bool = False) -> List[str]:
        """Provide to command to do the sync

        Args:
            fail_on_missing: Flag to make the command fail if any of
                the sources are missing.

        Returns:
            The command to do the sync, as a list of strings.
        """
        command = ["rsync", "--archive"]
        if self.mkpath:
            command += ["--mkpath"]
        if not fail_on_missing:
            command += ["--ignore-missing-args"]
        for src in self.src:
            globs = glob.glob(str(src))
            if globs:
                command += globs
            else:
                # 'globs' is an empty list if 'src' is missing, which could
                # render the 'rsync' command meaningless.
                # Make sure 'src' is part of the command in these cases,
                # and let --ignore-missing-args handle the rest.
                command += [str(src)]
        command += [str(self.dest)]
        return command

    def resolve(self, src_base: Path = Path.cwd(), dest_base: Path = Path.cwd()):
        """Resolve all paths and check they are relative to src_base and dest_base

        Args:
            src_base: Base directory for all src items.
            dest_base: Base directory for dest.
        """
        self.src = [check_subpath(src_base / path, src_base) for path in self.src]
        self.dest = check_subpath(dest_base / self.dest, dest_base)

    def drop_src(
        self, src: Union[str, Path], criteria=lambda x, y: x == Path(y)
    ) -> Optional["SyncFilesItem"]:
        """Remove 'src' from the list of src-s

        Args:
            src: A path to be removed.
            criteria: Function to tell if a src should be removed.
                Receives two arguments: the src item inspected and
                'src' received by this method.
        Returns:
            A *new* SyncFilesItem instance if the internal src list
            has still some items left after 'src' is removed.
            Otherwise returns None.
        """
        new_src = [s for s in self.src if not criteria(s, src)]
        if new_src:
            return SyncFilesItem(new_src, self.dest)
        else:
            return None


def iter_srcs(synced_files: Sequence[SyncFilesItem]):
    """Iterate over all the src-s in a list of SyncFilesItem

    Args:
        synced_files: List of SyncFilesItem.

    Yields:
        src-s from every SyncFilesItem, one by one.
    """
    for item in synced_files:
        yield from item.src


def sync_files(synced_files: Sequence[SyncFilesItem]):
    """
    Copy files b/w upstream and downstream repo.
    """
    for item in synced_files:
        run_command(item.command(), print_live=True)



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
E2E tests which utilize cockpit projects
"""

import pytest

from packit.cli.utils import get_packit_api
from packit.local_project import LocalProject
from packit.utils.commands import cwd
from tests.spellbook import (
    initiate_git_repo,
    get_test_config,
    UP_SNAPD,
    UP_OSBUILD,
    UP_EDD,
    DG_OGR,
    build_srpm,
    UP_VSFTPD,
)


@pytest.fixture(
    params=[
        (UP_SNAPD, "2.41", "https://github.com/snapcore/snapd"),
        (UP_OSBUILD, "2", "https://github.com/osbuild/osbuild"),
        (UP_EDD, "0.3", "https://github.com/psss/edd"),
        (UP_VSFTPD, "3.0.3", "https://github.com/olysonek/vsftpd"),
        (DG_OGR, None, "https://src.fedoraproject.org/rpms/python-ogr"),
    ]
)
def example_repo(request, tmp_path):
    example_path, tag, remote = request.param
    u = tmp_path / "up"
    initiate_git_repo(u, tag=tag, copy_from=example_path, upstream_remote=remote)
    return u


def test_srpm_on_example(example_repo):
    c = get_test_config()
    api = get_packit_api(config=c, local_project=LocalProject(working_dir=example_repo))
    with cwd(example_repo):
        path = api.create_srpm()
    assert path.exists()
    build_srpm(path)



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import subprocess

import pytest
from ogr.services.pagure import PagureService

from tests.spellbook import git_set_user_email


@pytest.mark.skipif(
    condition=True, reason="Don't interact with a real pagure instance by default"
)
def test_basic_distgit_workflow(tmp_path):
    pagure_token = os.getenv("PAGURE_TOKEN")

    pag = PagureService(
        token=pagure_token,
        repo="tmux-top",
        namespace="rpms",
        instance_url="https://src.stg.fedoraproject.org/",
    )

    print(pag.pagure.whoami())

    proj = pag.get_project()

    fork = proj.get_fork(create=True)
    clone_url = fork.get_git_urls()["ssh"]

    repo = tmp_path.joinpath("repo")

    branch_name = "cookie"

    subprocess.check_call(["git", "clone", clone_url, repo])
    git_set_user_email(repo)
    subprocess.check_call(["git", "checkout", "-B", branch_name], cwd=repo)

    repo.joinpath("README").write_text("just trying something out\n")

    subprocess.check_call(["git", "add", "README"], cwd=repo)
    subprocess.check_call(["git", "commit", "-m", "test commit"], cwd=repo)
    subprocess.check_call(
        ["git", "push", "origin", f"{branch_name}:{branch_name}"], cwd=repo
    )

    pr = fork.create_pr("testing PR", "serious description", "master", branch_name)

    proj.pr_comment(pr.id, "howdy!")



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import shutil
from pathlib import Path
from typing import Tuple
import subprocess
import tempfile

import pytest
from flexmock import flexmock

from packit.utils.commands import cwd
from packit.utils.repo import create_new_repo
from tests.spellbook import (
    initiate_git_repo,
    UPSTREAM,
    prepare_dist_git_repo,
    DISTGIT,
    DG_OGR,
    UPSTREAM_SPEC_NOT_IN_ROOT,
    UPSTREAM_WITH_MUTLIPLE_SOURCES,
    UPSTREAM_WEIRD_SOURCES,
)


# define own tmp_path fixture for older version of pytest (Centos)
try:
    from _pytest import tmpdir

    _ = tmpdir.tmp_path
except (ImportError, AttributeError, KeyError):

    @pytest.fixture()
    def tmp_path():
        TMP_DIR = "/tmp/pytest_tmp_path/"
        Path(TMP_DIR).mkdir(exist_ok=True, parents=True)
        return Path(tempfile.mkdtemp(prefix=TMP_DIR))


def get_git_repo_and_remote(
    target_dir: Path, repo_template_path: Path
) -> Tuple[Path, Path]:
    """
    :param target_dir: tmpdir from pytest - we'll work here
    :param repo_template_path: git repo template from tests/data/
    """
    u_remote_path = target_dir / f"upstream_remote-{repo_template_path.name}"
    u_remote_path.mkdir(parents=True, exist_ok=True)
    create_new_repo(u_remote_path, ["--bare"])

    u = target_dir / f"local_clone-{repo_template_path.name}"
    shutil.copytree(repo_template_path, u)
    initiate_git_repo(u, tag="0.1.0", push=True, upstream_remote=str(u_remote_path))

    return u, u_remote_path


@pytest.fixture()
def upstream_and_remote(tmp_path) -> Tuple[Path, Path]:
    return get_git_repo_and_remote(tmp_path, UPSTREAM)


@pytest.fixture()
def upstream_and_remote_with_multiple_sources(tmp_path) -> Tuple[Path, Path]:
    return get_git_repo_and_remote(tmp_path, UPSTREAM_WITH_MUTLIPLE_SOURCES)


@pytest.fixture()
def upstream_and_remote_weird_sources(tmp_path) -> Tuple[Path, Path]:
    return get_git_repo_and_remote(tmp_path, UPSTREAM_WEIRD_SOURCES)


@pytest.fixture()
def upstream_spec_not_in_root(tmp_path) -> Tuple[Path, Path]:
    return get_git_repo_and_remote(tmp_path, UPSTREAM_SPEC_NOT_IN_ROOT)


@pytest.fixture()
def distgit_and_remote(tmp_path) -> Tuple[Path, Path]:
    d_remote_path = tmp_path / "dist_git_remote"
    d_remote_path.mkdir(parents=True, exist_ok=True)
    create_new_repo(d_remote_path, ["--bare"])

    d = tmp_path / "dist_git"
    shutil.copytree(DISTGIT, d)
    initiate_git_repo(
        d,
        push=True,
        remotes=[
            ("origin", str(d_remote_path)),
            ("i_am_distgit", "https://src.fedoraproject.org/rpms/python-ogr"),
        ],
    )
    prepare_dist_git_repo(d)

    return d, d_remote_path


@pytest.fixture()
def ogr_distgit_and_remote(tmp_path) -> Tuple[Path, Path]:
    d_remote_path = tmp_path / "ogr_dist_git_remote"
    d_remote_path.mkdir(parents=True, exist_ok=True)
    create_new_repo(d_remote_path, ["--bare"])

    d = tmp_path / "ogr_dist_git"
    shutil.copytree(DG_OGR, d)
    initiate_git_repo(
        d,
        push=True,
        remotes=[
            ("origin", str(d_remote_path)),
            ("i_am_distgit", "https://src.fedoraproject.org/rpms/python-ogr"),
        ],
    )
    prepare_dist_git_repo(d)
    return d, d_remote_path


@pytest.fixture(params=["upstream", "ogr-distgit"])
def upstream_or_distgit_path(
    request, upstream_and_remote, distgit_and_remote, ogr_distgit_and_remote
):
    """
    Parametrize the test to upstream, downstream [currently skipped] and ogr distgit
    """
    cwd_path = {
        "upstream": upstream_and_remote[0],
        "distgit": distgit_and_remote[0],
        "ogr-distgit": ogr_distgit_and_remote[0],
    }[request.param]

    return cwd_path


@pytest.fixture(
    params=["upstream", "distgit", "ogr-distgit", "upstream-with-multiple-sources"]
)
def cwd_upstream_or_distgit(
    request,
    upstream_and_remote,
    distgit_and_remote,
    ogr_distgit_and_remote,
    upstream_and_remote_with_multiple_sources,
):
    """
    Run the code from upstream, downstream and ogr-distgit.

    When using be careful to
        - specify this fixture in the right place
        (the order of the parameters means order of the execution)
        - to not overwrite the cwd in the other fixture or in the test itself
    """
    cwd_path = {
        "upstream": upstream_and_remote[0],
        "distgit": distgit_and_remote[0],
        "ogr-distgit": ogr_distgit_and_remote[0],
        "upstream-with-multiple-sources": upstream_and_remote_with_multiple_sources[0],
    }[request.param]

    with cwd(cwd_path):
        yield cwd_path


@pytest.fixture
def copr_client_mock(get_list_return=None):
    get_list_return = {
        "centos-stream-aarch64": "",
        "custom-1-x86_64": "",
        "epel-6-x86_64": "",
        "epel-8-x86_64": "",
        "fedora-31-aarch64": "",
        "fedora-31-armhfp": "This is emulated on x86_64",
        "fedora-31-x86_64": "",
        "fedora-32-armhfp": "This is emulated on x86_64",
        "fedora-32-i386": "Not-released Koji packages",
        "fedora-32-s390x": "This is emulated on x86_64",
        "fedora-32-x86_64": "",
        "fedora-33-x86_64": "",
        "fedora-eln-x86_64": "",
        "fedora-rawhide-aarch64": "",
        "fedora-rawhide-x86_64": "",
    }

    copr_mock = flexmock(mock_chroot_proxy=flexmock(get_list=lambda: get_list_return))
    return copr_mock


@pytest.fixture(autouse=True, scope="function")
def configure_git():
    CMDS = [
        ["git", "config", "--global", "user.email", "packit@redhat.com"],
        ["git", "config", "--global", "user.name", "Packit Test"],
    ]
    # verify that git is already configured
    try:
        output = subprocess.check_output(["git", "config", "-l", "--global"])
    except subprocess.CalledProcessError:
        output = ""
    if "user.name" in output if isinstance(output, str) else output.decode():
        return
    for item in CMDS:
        subprocess.call(item)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import os
import shlex
import subprocess
import sys
from contextlib import contextmanager
from pathlib import Path
from typing import Dict, List, Optional, Union

from packit.exceptions import PackitCommandFailedError
from packit.utils.logging import StreamLogger

logger = logging.getLogger(__name__)


def run_command(
    cmd: Union[List[str], str],
    error_message: str = None,
    cwd: Union[str, Path] = None,
    fail: bool = True,
    output: bool = False,
    env: Optional[Dict] = None,
    decode: bool = True,
    print_live: bool = False,
):
    """
    run provided command in a new subprocess

    :param cmd: 'duh
    :param error_message: if the command fails, output this error message
    :param cwd: run the command in
    :param fail: raise an exception when the command fails
    :param output: if True, return command's stdout & stderr
    :param env: set these env vars in the subprocess
    :param decode: decode stdout from utf8 to string
    :param print_live: print output from the command realtime to INFO log
    """
    if not isinstance(cmd, list):
        cmd = shlex.split(cmd)

    escaped_command = " ".join(cmd)

    logger.debug(f"Command: {escaped_command}")

    cwd = str(cwd) if cwd else str(Path.cwd())
    error_message = error_message or f"Command {escaped_command!r} failed."

    # we need to pass complete env to Popen, otherwise we lose everything from os.environ
    cmd_env = os.environ
    if env:
        cmd_env.update(env)

    # we can't use universal newlines here b/c the output from the command can be encoded
    # in something alien and we would "can't decode this using utf-8" errors
    # https://github.com/packit/systemd-rhel8-flock/pull/9#issuecomment-550184016
    shell = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=False,
        cwd=cwd,
        env=cmd_env,
    )

    stdout = StreamLogger(
        shell.stdout,
        log_level=logging.DEBUG if not print_live else logging.INFO,
        decode=decode,
    )
    stderr = StreamLogger(
        shell.stderr,
        log_level=logging.DEBUG if not print_live else logging.INFO,
        decode=decode,
    )

    stdout.start()
    stderr.start()
    shell.wait()
    stdout.join()
    stderr.join()

    if shell.returncode != 0:
        logger.error(f"{error_message}")
        if fail:
            stderr_output = (
                stderr.get_output().decode() if decode else stderr.get_output()
            )
            stdout_output = (
                stdout.get_output().decode() if decode else stdout.get_output()
            )
            if output:
                logger.debug(f"Command stderr: {stderr_output}")
                logger.debug(f"Command stdout: {stdout_output}")
            raise PackitCommandFailedError(
                f"{error_message}",
                stdout_output=stdout_output,
                stderr_output=stderr_output,
            )
        success = False
    else:
        success = True

    if not output:
        return success

    o = stdout.get_output()
    return o.decode(sys.getdefaultencoding()) if decode else o


def run_command_remote(
    cmd,
    error_message=None,
    cwd=None,
    fail=True,
    output=False,
    env: Optional[Dict] = None,
    decode=True,
    print_live: bool = False,
):
    """
    wrapper for run_command method
    Indicating that this command run some action without local effect,
    or the effect is not important.

    eg.
        submit something to some server, and check how server reply
        call kinit of some ticket
    """
    return run_command(
        cmd, error_message, cwd, fail, output, env, decode=decode, print_live=print_live
    )


@contextmanager
def cwd(target: Union[str, Path]):
    """
    Manage cwd in a pushd/popd fashion.

    Usage:

        with cwd(tmpdir):
          do something in tmpdir
    """
    curdir = os.getcwd()
    os.chdir(target)
    try:
        yield
    finally:
        os.chdir(curdir)



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from pathlib import Path
from typing import Optional, Union

from packit.exceptions import PackitCommandFailedError

from packit.utils import commands  # so we can mock utils
from packit.utils.logging import logger


class FedPKG:
    """
    Part of the code is from release-bot:

    https://github.com/user-cont/release-bot/blob/master/release_bot/fedora.py
    """

    def __init__(
        self,
        fas_username: str = None,
        directory: Union[Path, str] = None,
        fedpkg_exec: str = "fedpkg",
    ):
        self.fas_username = fas_username
        self.directory = Path(directory) if directory else None
        self.fedpkg_exec = fedpkg_exec

    def __repr__(self):
        return (
            "FedPKG("
            f"fas_username='{self.fas_username}', "
            f"directory='{self.directory}', "
            f"fedpkg_exec='{self.fedpkg_exec}')"
        )

    def new_sources(self, sources="", fail=True):
        if not self.directory.is_dir():
            raise Exception("Cannot access fedpkg repository:")

        return commands.run_command_remote(
            cmd=[self.fedpkg_exec, "new-sources", sources],
            cwd=self.directory,
            error_message="Adding new sources failed:",
            fail=fail,
        )

    def build(
        self,
        scratch: bool = False,
        nowait: bool = False,
        koji_target: Optional[str] = None,
        srpm_path: Optional[Path] = None,
    ):
        """
        build in koji

        :param scratch: scratch (temporary) build or not?
        :param nowait: False == wait for the build to finish
        :param koji_target: koji target to build in (`koji list-targets`)
        :param srpm_path: use selected SRPM for build, not dist-git repo & ref
        :return:
        """
        cmd = [self.fedpkg_exec, "build"]
        if scratch:
            cmd.append("--scratch")
        if nowait:
            cmd.append("--nowait")
        if koji_target:
            cmd += ["--target", koji_target]
        if srpm_path:
            cmd += ["--srpm", str(srpm_path)]

        try:
            commands.run_command_remote(
                cmd=cmd,
                cwd=self.directory,
                error_message="Submission of build to koji failed.",
                fail=True,
            )

        except PackitCommandFailedError as ex:
            # fail on the fedpkg side, the build is triggered
            if (
                "watch_tasks() got an unexpected keyword argument 'ki_handler'"
                in ex.stderr_output
            ):
                logger.info(
                    "The 'fedpkg build' command crashed which is a known issue: "
                    "the build is submitted in koji anyway."
                )
                logger.debug(ex.stdout_output)

            else:
                raise

    def clone(self, package_name: str, target_path: str, anonymous: bool = False):
        """
        clone a dist-git repo; this has to be done in current env
        b/c we don't have the keytab in sandbox
        """
        cmd = [self.fedpkg_exec]
        if self.fas_username:
            cmd += ["--user", self.fas_username]
        cmd += ["-q", "clone"]
        if anonymous:
            cmd += ["-a"]
        cmd += [package_name, target_path]

        error_msg = (
            f"Packit failed to clone the repository {package_name}; "
            "please make sure that you are authorized to clone repositories "
            "from Fedora dist-git - this may require SSH keys set up or "
            "Kerberos ticket being active."
        )
        commands.run_command(cmd=cmd, error_message=error_msg)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from pathlib import Path

import pytest

from packit.api import PackitAPI
from packit.utils.commands import cwd


@pytest.mark.parametrize(
    "raw_package_config,expected_output",
    [
        (
            """
            {
                "config_file_path": "packit.json",
                "dist_git_base_url": "https://packit.dev/",
                "downstream_package_name": "packit",
                "upstream_ref": "last_commit",
                "upstream_package_name": "packit_upstream",
                "create_tarball_command": ["commands"],
                "allowed_gpg_keys": ["gpg"],
                "dist_git_namespace": "awesome",
                "notifications": {
                    "pull_request": {
                        "successful_build": True
                    }
                }
            }
            """,
            "packit.json is valid and ready to be used",
        ),
        (
            """
            {
                "config_file_path": "packit.json",
                "dist_git_base_url": "https://packit.dev/",
                "downstream_package_name": "packit",
                "upstream_ref": "last_commit",
                "upstream_package_name": "packit_upstream",
                "create_tarball_command": ["commands"],
                "allowed_gpg_keys": ["gpg"],
                "dist_git_namespace": "awesome",
                "notifications": {
                    "pull_request": {
                        "successful_build": 55
                    }
                }
            }
            """,
            "* field notifications has an incorrect value:\n"
            "** field pull_request has an incorrect value:\n"
            "*** value at index successful_build: Not a valid boolean.",
        ),
        ("{}", "packit.json is valid and ready to be used"),
        (
            """
            {
                "config_file_path": "packit.json",
                "dist_git_base_url": "https://packit.dev/",
                "downstream_package_name": "packit",
                "upstream_ref": "last_commit",
                "upstream_package_name": "packit_upstream",
                "create_tarball_command": ["commands"],
                "allowed_gpg_keys": ["gpg"],
                "dist_git_namespace": "awesome",
                "synced_files": ["a.md", "b.md", "c.txt"]
            }
            """,
            "packit.json is valid and ready to be used",
        ),
        (
            """
                {
                    "config_file_path": "packit.json",
                    "dist_git_base_url": "https://packit.dev/",
                    "downstream_package_name": "packit",
                    "upstream_ref": "last_commit",
                    "upstream_package_name": "packit_upstream",
                    "create_tarball_command": ["commands"],
                    "allowed_gpg_keys": ["gpg"],
                    "dist_git_namespace": "awesome",
                    "synced_files": [{ "src": 55, "dest": "a.md" }, "b.md", "c.txt"]
                }
                """,
            "Expected 'list[str]' or 'str', got <class 'int'>.",
        ),
        (
            """
                {
                    "config_file_path": "packit.json",
                    "dist_git_base_url": "https://packit.dev/",
                    "downstream_package_name": "packit",
                    "upstream_ref": "last_commit",
                    "upstream_package_name": "packit_upstream",
                    "create_tarball_command": ["commands"],
                    "allowed_gpg_keys": ["gpg"],
                    "dist_git_namespace": "awesome",
                    "synced_files": ["a.md", "b.md", { "src": "c.txt", "dest": True }]
                }
                """,
            "dest: Not a valid string.",
        ),
        (
            """
                {
                    "dist_git_base_url": "https://packit.dev/",
                    "downstream_package_name": "packit",
                    "upstream_ref": "last_commit",
                    "upstream_package_name": "packit_upstream",
                    "create_tarball_command": ["commands"],
                    "allowed_gpg_keys": ["gpg"],
                    "dist_git_namespace": "awesome"
                }
                """,
            "packit.json is valid and ready to be used",
        ),
        (
            """
                {
                    "dist_git_base_url": "https://packit.dev/",
                    "downstream_package_name": 23,
                    "upstream_ref": "last_commit",
                    "upstream_package_name": "packit_upstream",
                    "create_tarball_command": ["commands"],
                    "allowed_gpg_keys": ["gpg"],
                    "dist_git_namespace": "awesome"
                }
                """,
            "* field downstream_package_name: Not a valid string.",
        ),
        (
            """
                {
                    "dist_git_base_url": "https://packit.dev/",
                    "downstream_package_name": "packit",
                    "upstream_ref": "last_commit",
                    "upstream_package_name": "packit_upstream",
                    "create_tarball_command": ["commands"],
                    "allowed_gpg_keys": ["gpg"],
                    "dist_git_namespace": "awesome",
                    create_pr: ""
                }
                """,
            "* field create_pr: Not a valid boolean.",
        ),
        (
            """
                {
                    "config_file_path": "packit.json",
                    "dist_git_base_url": "https: //packit.dev/",
                    "downstream_package_name": "packit",
                    "upstream_ref": "last_commit",
                    "upstream_package_name": "packit_upstream",
                    "create_tarball_command": ["commands"],
                    "allowed_gpg_keys": ["gpg"],
                    "dist_git_namespace": "awesome"
                }
                """,
            "packit.json is valid and ready to be used",
        ),
        (
            """
                {
                    "config_file_path": "packit.json",
                    "dist_git_base_url": "https: //packit.dev/",
                    "downstream_package_name": "packit",
                    "upstream_ref": "last_commit",
                    "upstream_package_name": "packit_upstream",
                    "create_tarball_command": ["commands"],
                    "allowed_gpg_keys": "gpg",
                    "dist_git_namespace": "awesome"
                }
                """,
            "* field allowed_gpg_keys: Not a valid list.",
        ),
        (
            """
                {
                    "config_file_path":"packit.json",
                    "dist_git_base_url":"https://packit.dev/",
                    "downstream_package_name":"packit",
                    "upstream_ref":"last_commit",
                    "upstream_package_name":"packit_upstream",
                    "create_tarball_command":[25],
                    "allowed_gpg_keys":["gpg"],
                    "dist_git_namespace":"awesome"
                }
                """,
            "* field create_tarball_command has an incorrect value:\n"
            "** value at index 0: Not a valid string.",
        ),
        (
            """
                {
                    "config_file_path":"packit.json",
                    "dist_git_base_url":"https://packit.dev/",
                    "downstream_package_name":"packit",
                    "upstream_ref":"last_commit",
                    "upstream_package_name":"packit_upstream",
                    "create_tarball_command":["commands", True],
                    "allowed_gpg_keys":["gpg"],
                    "dist_git_namespace":"awesome"
                }
                """,
            "* field create_tarball_command has an incorrect value:\n"
            "** value at index 1: Not a valid string.",
        ),
    ],
    ids=[
        "valid_1",
        "notif_succ_build",
        "empty",
        "valid_2",
        "synced_files_src",
        "synced_files_dest",
        "valid_3",
        "downstream_name",
        "create_pr",
        "valid_4",
        "allowed_gpg",
        "create_tarball_1",
        "create_tarball_2",
    ],
)
def test_schema_validation(tmpdir, raw_package_config, expected_output):
    with cwd(tmpdir):
        Path("packit.json").write_text(raw_package_config)
        Path("packit.spec").write_text("hello")

        output = PackitAPI.validate_package_config(Path("."))
        assert expected_output in output



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from subprocess import check_output
import unittest
from flexmock import flexmock
import rebasehelper
from rebasehelper.exceptions import RebaseHelperError
from requre.cassette import DataTypes
from requre.online_replacing import (
    apply_decorator_to_all_methods,
    record_requests_for_all_methods,
    replace_module_match,
)
from requre.helpers.tempfile import TempFile
from requre.helpers.simple_object import Simple
from requre.helpers.files import StoreFiles
from requre.helpers.git.pushinfo import PushInfoStorageList
from requre.helpers.git.fetchinfo import FetchInfoStorageList
from requre.helpers.git.repo import Repo

from packit.api import PackitAPI
from tests_recording.testbase import PackitTest


@record_requests_for_all_methods()
@apply_decorator_to_all_methods(
    replace_module_match(
        what="packit.utils.run_command_remote", decorate=Simple.decorator_plain()
    )
)
@apply_decorator_to_all_methods(
    replace_module_match(
        what="packit.fedpkg.FedPKG.clone",
        decorate=StoreFiles.where_arg_references(
            key_position_params_dict={"target_path": 2}
        ),
    )
)
@apply_decorator_to_all_methods(
    replace_module_match(
        what="git.repo.base.Repo.clone_from",
        decorate=StoreFiles.where_arg_references(
            key_position_params_dict={"to_path": 2},
            return_decorator=Repo.decorator_plain,
        ),
    )
)
@apply_decorator_to_all_methods(
    replace_module_match(
        what="git.remote.Remote.push", decorate=PushInfoStorageList.decorator_plain()
    )
)
@apply_decorator_to_all_methods(
    replace_module_match(
        what="git.remote.Remote.fetch", decorate=FetchInfoStorageList.decorator_plain()
    )
)
@apply_decorator_to_all_methods(
    replace_module_match(what="tempfile.mkdtemp", decorate=TempFile.mkdtemp())
)
@apply_decorator_to_all_methods(
    replace_module_match(what="tempfile.mktemp", decorate=TempFile.mktemp())
)
# Be aware that decorator stores login and token to test_data, replace it by some value.
# Default precommit hook doesn't do that for copr.v3.helpers, see README.md
@apply_decorator_to_all_methods(
    replace_module_match(
        what="copr.v3.helpers.config_from_file", decorate=Simple.decorator_plain()
    )
)
class ProposeUpdate(PackitTest):
    def cassette_setup(self, cassette):
        cassette.data_miner.data_type = DataTypes.Dict

    def setUp(self):
        super().setUp()
        self.set_git_user()
        self._api = None

    @property
    def api(self):
        if not self._api:
            self._api = PackitAPI(
                config=self.config,
                package_config=self.pc,
                upstream_local_project=self.lp,
            )
            self.api._up = self.upstream
            self.api._dg = self.dg
        return self._api

    def check_version_increase(self):
        """Bump version in specfile and tag it with the new version.
        Might fail if such tag already exists in requre repo.
        In that case you probably need to bump Version in fedora/python-requre.spec
        """
        filedata = self.project_specfile_location.read_text()
        # Patch the specfile with new version
        version_increase = "0.0.0"
        for line in filedata.splitlines():
            if "Version:" in line:
                version = line.rsplit(" ", 1)[1]
                v1, v2, v3 = version.split(".")
                version_increase = ".".join([v1, str(int(v2) + 1), v3])
                filedata = filedata.replace(version, version_increase)
                break
        self.project_specfile_location.write_text(filedata)
        check_output(
            f"cd {self.lp.working_dir};"
            f"git commit -m 'test change' {self._project_specfile_path.name};"
            f"git tag -a {version_increase} -m 'my version {version_increase}'",
            shell=True,
        )
        self.api.sync_release(force=True)

    @unittest.skip("Recordings are broken")
    def test_comment_in_spec(self):
        """
        change specfile little bit to have there some change, do not increase version
        """
        with self.project_specfile_location.open("a") as myfile:
            myfile.write("\n# comment\n")
        version_increase = "10.0.0"
        check_output(
            f"cd {self.lp.working_dir};"
            f"git commit -m 'test change' {self._project_specfile_path};"
            f"git tag -a {version_increase} -m 'my version {version_increase}'",
            shell=True,
        )
        self.api.sync_release()

    @unittest.skip("Recordings are broken")
    def test_changelog_sync(self):
        """
        Bump version two times and see if the changelog is synced
        when it's configured to do so.
        """
        original_upstream_spec_content = self.api.up.absolute_specfile_path.read_text()
        check_output(
            f"cd {self.lp.working_dir};"
            f"rpmdev-bumpspec {self._project_specfile_path};"
            f"rpmdev-bumpspec {self._project_specfile_path};"
            f"git commit -m 'test change' {self._project_specfile_path}",
            shell=True,
        )
        changed_upstream_spec_content = self.api.up.absolute_specfile_path.read_text()
        assert original_upstream_spec_content != changed_upstream_spec_content
        self.api.package_config.sync_changelog = True
        self.api.sync_release(use_local_content=True)
        new_downstream_spec_content = self.api.dg.absolute_specfile_path.read_text()
        assert changed_upstream_spec_content == new_downstream_spec_content

    @unittest.skipIf(
        int(getattr(rebasehelper, "VERSION", "0.18.0").split(".")[1]) >= 24,
        "Old version of RebaseHelper raises exception",
    )
    def test_version_change_exception(self):
        """
        check if it raises exception, because sources are not uploaded in distgit
        Downgrade rebasehelper to version < 0.24.0
        """
        self.assertRaises(RebaseHelperError, self.check_version_increase)

    @unittest.skipIf(
        int(getattr(rebasehelper, "VERSION", "0.18.0").split(".")[1]) < 24,
        "New version of rebasehelper >=24",
    )
    @unittest.skip("Recordings are broken")
    def test_version_change_exception_new_rebase_helper(self):
        """
        check if it raises exception, because sources are not uploaded in distgit
        New version raises same error, but it raises it on another place
        Upgrade rebase helper >= 0.24.0
        """
        self.assertRaises(RebaseHelperError, self.check_version_increase)

    @unittest.skip("Recordings are broken")
    def test_version_change_mocked(self):
        """
        version is not not uploaded, so skip in this test
        """
        flexmock(self.api).should_receive("_handle_sources").and_return(None)
        self.check_version_increase()



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from pathlib import Path
from typing import Union, List, Optional, Dict

from flexmock import flexmock

from packit.utils import commands


def test_basic_build(
    cwd_upstream_or_distgit, api_instance, mock_remote_functionality_upstream
):
    u, d, api = api_instance
    flexmock(api).should_receive("init_kerberos_ticket").at_least().once()
    flexmock(commands).should_receive("run_command_remote").with_args(
        cmd=["fedpkg", "build", "--scratch", "--nowait", "--target", "asdqwe"],
        cwd=api.dg.local_project.working_dir,
        error_message="Submission of build to koji failed.",
        fail=True,
    ).once()

    api.build("main", scratch=True, nowait=True, koji_target="asdqwe")


def test_build_from_upstream(
    cwd_upstream_or_distgit, api_instance, mock_remote_functionality_upstream
):
    u, d, api = api_instance

    def mocked_run_command(
        cmd: Union[List[str], str],
        error_message: str = None,
        cwd: Union[str, Path] = None,
        fail: bool = True,
        output: bool = False,
        env: Optional[Dict] = None,
        decode=True,
        print_live=False,
    ):
        assert cmd[:-1] == ["koji", "build", "--scratch", "--nowait", "hadron-collider"]
        srpm_path = cmd[-1]
        assert Path(srpm_path).is_file()
        assert srpm_path.endswith(".src.rpm")
        assert cwd == api.up.local_project.working_dir
        return "\n\nLink to koji build: https://koji...\n"

    flexmock(commands, run_command_remote=mocked_run_command)
    api.build(
        "master",
        scratch=True,
        nowait=True,
        from_upstream=True,
        koji_target="hadron-collider",
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
This is the official python interface for packit.
"""

import asyncio
import logging
import shutil
import sys
from datetime import datetime
from pathlib import Path
from typing import Sequence, Callable, List, Tuple, Dict, Iterable, Optional, Union

from ogr.abstract import PullRequest
from pkg_resources import get_distribution, DistributionNotFound
from tabulate import tabulate

from packit.actions import ActionName
from packit.config import Config
from packit.config.common_package_config import CommonPackageConfig
from packit.config.package_config import find_packit_yaml, load_packit_yaml
from packit.config.package_config_validator import PackageConfigValidator
from packit.constants import SYNCING_NOTE
from packit.copr_helper import CoprHelper
from packit.distgit import DistGit
from packit.exceptions import (
    PackitException,
    PackitSRPMException,
    PackitSRPMNotFoundException,
    PackitRPMException,
    PackitRPMNotFoundException,
    PackitCoprException,
)
from packit.local_project import LocalProject
from packit.patches import PatchGenerator
from packit.source_git import SourceGitGenerator
from packit.status import Status
from packit.sync import sync_files, SyncFilesItem
from packit.upstream import Upstream
from packit.utils import commands
from packit.utils.extensions import assert_existence

logger = logging.getLogger(__name__)


def get_packit_version() -> str:
    try:
        return get_distribution("packitos").version
    except DistributionNotFound:
        return "NOT_INSTALLED"


class PackitAPI:
    def __init__(
        self,
        config: Config,
        package_config: Optional[
            CommonPackageConfig
        ],  # validate doesn't want PackageConfig
        upstream_local_project: LocalProject = None,
        downstream_local_project: LocalProject = None,
        stage: bool = False,
    ) -> None:
        self.config = config
        self.package_config: CommonPackageConfig = package_config
        self.upstream_local_project = upstream_local_project
        self.downstream_local_project = downstream_local_project
        self.stage = stage

        self._up = None
        self._dg = None
        self._copr_helper: Optional[CoprHelper] = None
        self._kerberos_initialized = False

    def __repr__(self):
        return (
            "PackitAPI("
            f"config='{self.config}', "
            f"package_config='{self.package_config}', "
            f"upstream_local_project='{self.upstream_local_project}', "
            f"downstream_local_project='{self.downstream_local_project}', "
            f"up='{self.up}', "
            f"dg='{self.dg}', "
            f"copr_helper='{self.copr_helper}', "
            f"stage='{self.stage}')"
        )

    @property
    def up(self):
        if self._up is None:
            self._up = Upstream(
                config=self.config,
                package_config=self.package_config,
                local_project=self.upstream_local_project,
            )
        return self._up

    @property
    def dg(self):
        if self._dg is None:
            self.init_kerberos_ticket()
            self._dg = DistGit(
                config=self.config,
                package_config=self.package_config,
                local_project=self.downstream_local_project,
            )
        return self._dg

    @property
    def copr_helper(self) -> CoprHelper:
        if self._copr_helper is None:
            self._copr_helper = CoprHelper(
                upstream_local_project=self.upstream_local_project
            )
        return self._copr_helper

    def update_dist_git(
        self,
        version: Optional[str],
        upstream_ref: Optional[str],
        force_new_sources: bool,
        upstream_tag: str,
        commit_msg: str,
    ):
        """Update a dist-git repo from an upstream (aka source-git) repo

        - copy files to be synced to dist-git
        - generate and update patch files and the spec-file
        - upload source archives to the lookaside cache

        Args:
            version: Upstream version to update in Fedora.
            upstream_ref: For a source-git repo, use this ref as the latest upstream commit.
            force_new_sources: Don't check the lookaside cache and perform new-sources.
            upstream_tag: Use the message of the commit referenced by this tag to update the
                changelog in the spec-file, if requested.
            commit_msg: Use this commit message in dist-git.
        """
        synced_files = self.package_config.get_all_files_to_sync()
        # Make all paths absolute and check that they are within
        # the working directories of the repositories.
        for item in synced_files:
            item.resolve(
                src_base=self.up.local_project.working_dir,
                dest_base=self.dg.local_project.working_dir,
            )

        if self.up.with_action(action=ActionName.prepare_files):
            synced_files = self._prepare_files_to_sync(
                synced_files=synced_files,
                full_version=version,
                upstream_tag=upstream_tag,
            )

        sync_files(synced_files)

        if upstream_ref and self.up.with_action(action=ActionName.create_patches):
            patches = self.up.create_patches(
                upstream=upstream_ref,
                destination=str(self.dg.absolute_specfile_dir),
            )
            patches = PatchGenerator.undo_identical(
                patches, self.dg.local_project.git_repo
            )
            self.dg.specfile_add_patches(patches)

        self._handle_sources(add_new_sources=True, force_new_sources=force_new_sources)

        self.dg.commit(title=f"{version} upstream release", msg=commit_msg)

    def sync_release(
        self,
        dist_git_branch: Optional[str] = None,
        version: Optional[str] = None,
        tag: Optional[str] = None,
        use_local_content=False,
        force_new_sources=False,
        upstream_ref: Optional[str] = None,
        create_pr: bool = True,
        force: bool = False,
        create_sync_note: bool = True,
    ) -> Optional[PullRequest]:
        """
        Update given package in Fedora

        :param dist_git_branch: branch in dist-git, defaults to repo's default branch
        :param use_local_content: don't check out anything
        :param version: upstream version to update in Fedora
        :param tag: upstream git tag
        :param force_new_sources: don't check the lookaside cache and perform new-sources
        :param upstream_ref: for a source-git repo, use this ref as the latest upstream commit
        :param create_pr: create a pull request if set to True
        :param force: ignore changes in the git index
        :param create_sync_note: whether to create a note about the sync in the dist-git repo

        :return created PullRequest if create_pr is True, else None
        """
        dist_git_branch = (
            dist_git_branch or self.dg.local_project.git_project.default_branch
        )
        # process version and tag parameters
        if version and tag:
            raise PackitException(
                "Function parameters version and tag are mutually exclusive."
            )
        elif not tag:
            version = version or self.up.get_version()
            if not version:
                raise PackitException(
                    "Could not figure out version of latest upstream release."
                )
            upstream_tag = self.up.convert_version_to_tag(version)
        else:
            upstream_tag = tag
            version = self.up.get_version_from_tag(tag)

        assert_existence(self.up.local_project, "Upstream local project")
        assert_existence(self.dg.local_project, "Dist-git local project")
        if self.dg.is_dirty():
            raise PackitException(
                f"The distgit repository {self.dg.local_project.working_dir} is dirty."
                f"This is not supported."
            )
        if not force and self.up.is_dirty() and not use_local_content:
            raise PackitException(
                "The repository is dirty, will not discard the changes. Use --force to bypass."
            )
        # do not add anything between distgit clone and saving gpg keys!
        self.up.allowed_gpg_keys = self.dg.get_allowed_gpg_keys_from_downstream_config()

        upstream_ref = self.up._expand_git_ref(
            upstream_ref or self.package_config.upstream_ref
        )
        create_pr = create_pr and self.package_config.create_pr
        self.up.run_action(actions=ActionName.post_upstream_clone)

        current_up_branch = self.up.active_branch
        try:
            # we want to check out the tag only when local_content is not set
            # and it's an actual upstream repo and not source-git
            if upstream_ref:
                logger.info(
                    "We will not check out the upstream tag "
                    "because this is a source-git repo."
                )
            elif not use_local_content:
                self.up.local_project.checkout_release(upstream_tag)

            self.dg.check_last_commit()

            self.up.run_action(actions=ActionName.pre_sync)
            self.dg.create_branch(
                dist_git_branch,
                base=f"remotes/origin/{dist_git_branch}",
                setup_tracking=True,
            )

            # fetch and reset --hard upstream/$branch?
            logger.info(f"Using {dist_git_branch!r} dist-git branch.")
            self.dg.update_branch(dist_git_branch)
            self.dg.checkout_branch(dist_git_branch)

            if create_pr:
                local_pr_branch = f"{version}-{dist_git_branch}-update"
                self.dg.create_branch(local_pr_branch)
                self.dg.checkout_branch(local_pr_branch)

            if create_sync_note:
                readme_path = self.dg.local_project.working_dir / "README.packit"
                logger.debug(f"README: {readme_path}")
                readme_path.write_text(
                    SYNCING_NOTE.format(packit_version=get_packit_version())
                )

            description = (
                f"Upstream tag: {upstream_tag}\n"
                f"Upstream commit: {self.up.local_project.commit_hexsha}\n"
            )
            self.update_dist_git(
                version,
                upstream_ref,
                force_new_sources,
                upstream_tag,
                commit_msg=description,
            )

            new_pr = None
            if create_pr:
                title = f"Update to upstream release {version}"

                if not self.dg.pr_exists(title, description.rstrip(), dist_git_branch):
                    new_pr = self.push_and_create_pr(
                        pr_title=title,
                        pr_description=description,
                        dist_git_branch=dist_git_branch,
                    )
            else:
                self.dg.push(refspec=f"HEAD:{dist_git_branch}")
        finally:
            if not use_local_content and not upstream_ref:
                self.up.local_project.git_repo.git.checkout(current_up_branch)
            self.dg.refresh_specfile()
            self.dg.local_project.git_repo.git.reset("--hard", "HEAD")
        return new_pr

    def _prepare_files_to_sync(
        self, synced_files: List[SyncFilesItem], full_version: str, upstream_tag: str
    ) -> List[SyncFilesItem]:
        """Update the spec-file by setting the version and updating the changelog

        Skip everything if the changelog should be synced from upstream.

        Args:
            synced_files: A list of SyncFilesItem.
            full_version: Version to be set in the spec-file.
            upstream_tag: The commit message of this commit is going to be used
                to update the changelog in the spec-file.

        Returns:
            The list of synced files with the spec-file removed if it was updated.
        """
        if self.package_config.sync_changelog:
            return synced_files
        comment = (
            self.up.local_project.git_project.get_release(name=full_version).body
            if self.package_config.copy_upstream_release_description
            else self.up.get_commit_messages(
                after=self.up.get_last_tag(upstream_tag), before=upstream_tag
            )
        )
        try:
            self.dg.set_specfile_content(self.up.specfile, full_version, comment)
        except FileNotFoundError as ex:
            # no downstream spec file: this is either a mistake or
            # there is no spec file in dist-git yet, hence warning
            logger.warning(
                f"There is not spec file downstream: {ex}, copying the one from upstream."
            )
            shutil.copy2(
                self.up.absolute_specfile_path,
                self.dg.get_absolute_specfile_path(),
            )

        # exclude spec, we have special plans for it
        return list(
            filter(
                None, [x.drop_src(self.up.absolute_specfile_path) for x in synced_files]
            )
        )

    def sync_from_downstream(
        self,
        dist_git_branch: str = None,
        upstream_branch: str = None,
        no_pr: bool = False,
        fork: bool = True,
        remote_name: str = None,
        exclude_files: Iterable[str] = None,
        force: bool = False,
        sync_only_specfile: bool = False,
    ):
        """
        Sync content of Fedora dist-git repo back to upstream

        :param dist_git_branch: branch in dist-git, defaults to repo's default branch
        :param upstream_branch: upstream branch, defaults to repo's default branch
        :param no_pr: won't create a pull request if set to True
        :param fork: forks the project if set to True
        :param remote_name: name of remote where we should push; if None, try to find a ssh_url
        :param exclude_files: files that will be excluded from the sync
        :param force: ignore changes in the git index
        :param sync_only_specfile: whether to sync only content of specfile
        """
        exclude_files = exclude_files or []
        if not dist_git_branch:
            dist_git_branch = self.dg.local_project.git_project.default_branch
            logger.info(f"Dist-git branch not set, defaulting to {dist_git_branch!r}.")
        if not upstream_branch:
            upstream_branch = self.up.local_project.git_project.default_branch
            logger.info(f"Upstream branch not set, defaulting to {upstream_branch!r}.")
        logger.info(f"Upstream active branch: {self.up.active_branch}")

        if not force and self.up.is_dirty():
            raise PackitException(
                "The repository is dirty, will not discard the changes. Use --force to bypass."
            )
        self.dg.update_branch(dist_git_branch)
        self.dg.checkout_branch(dist_git_branch)

        logger.info(f"Using {dist_git_branch!r} dist-git branch.")

        if no_pr:
            self.up.checkout_branch(upstream_branch)
        else:
            local_pr_branch = f"{dist_git_branch}-downstream-sync"
            self.up.create_branch(local_pr_branch)
            self.up.checkout_branch(local_pr_branch)

        files = (
            [self.package_config.get_specfile_sync_files_item()]
            if sync_only_specfile
            else self.package_config.synced_files
        )

        # Drop files to be excluded from the sync.
        for ef in exclude_files:
            files = [
                f.drop_src(ef, criteria=lambda x, y: x.name == y)
                for f in files
                if f is not None
            ]
        # Make paths absolute and check if they are within the
        # working directories.
        for file in files:
            file.resolve(
                src_base=self.dg.local_project.working_dir,
                dest_base=self.up.local_project.working_dir,
            )
        sync_files(files)

        if not no_pr:
            description = f"Downstream commit: {self.dg.local_project.commit_hexsha}\n"

            commit_msg = f"Sync from downstream branch {dist_git_branch!r}"
            pr_title = f"Update from downstream branch {dist_git_branch!r}"

            self.up.commit(title=commit_msg, msg=description)

            # the branch may already be up, let's push forcefully
            source_branch, fork_username = self.up.push_to_fork(
                self.up.local_project.ref,
                fork=fork,
                force=True,
                remote_name=remote_name,
            )
            self.up.create_pull(
                pr_title,
                description,
                source_branch=source_branch,
                target_branch=upstream_branch,
                fork_username=fork_username,
            )

    def push_and_create_pr(
        self, pr_title: str, pr_description: str, dist_git_branch: str
    ) -> PullRequest:
        # the branch may already be up, let's push forcefully
        self.dg.push_to_fork(self.dg.local_project.ref, force=True)
        return self.dg.create_pull(
            pr_title,
            pr_description,
            source_branch=self.dg.local_project.ref,
            target_branch=dist_git_branch,
        )

    def _handle_sources(self, add_new_sources, force_new_sources):
        if not (add_new_sources or force_new_sources):
            return

        make_new_sources = False
        # btw this is really naive: the name could be the same but the hash can be different
        # TODO: we should do something when such situation happens
        if force_new_sources or not self.dg.is_archive_in_lookaside_cache(
            self.dg.upstream_archive_name
        ):
            make_new_sources = True
        else:
            sources_file = self.dg.local_project.working_dir / "sources"
            if self.dg.upstream_archive_name not in sources_file.read_text():
                make_new_sources = True
        if make_new_sources:
            archive = self.dg.download_upstream_archive()
            self.init_kerberos_ticket()
            self.dg.upload_to_lookaside_cache(str(archive))

    def build(
        self,
        dist_git_branch: str,
        scratch: bool = False,
        nowait: bool = False,
        koji_target: Optional[str] = None,
        from_upstream: bool = False,
    ):
        """
        Build component in Fedora infra (defaults to koji)

        :param dist_git_branch: ref in dist-git
        :param scratch: should the build be a scratch build?
        :param nowait: don't wait on build?
        :param koji_target: koji target to pick (see `koji list-targets`)
        :param from_upstream: build directly from the upstream checkout?
        """
        logger.info(f"Using {dist_git_branch!r} dist-git branch")
        self.init_kerberos_ticket()

        if from_upstream:
            srpm_path = self.create_srpm(srpm_dir=self.up.local_project.working_dir)
            return self.up.koji_build(
                scratch=scratch,
                nowait=nowait,
                koji_target=koji_target,
                srpm_path=srpm_path,
            )

        self.dg.create_branch(
            dist_git_branch,
            base=f"remotes/origin/{dist_git_branch}",
            setup_tracking=True,
        )

        self.dg.update_branch(dist_git_branch)
        self.dg.checkout_branch(dist_git_branch)

        self.dg.build(scratch=scratch, nowait=nowait, koji_target=koji_target)

    def create_update(
        self,
        dist_git_branch: str,
        update_type: str,
        update_notes: str,
        koji_builds: Sequence[str] = None,
    ):
        """
        Create bodhi update

        :param dist_git_branch: git ref
        :param update_type: type of the update, check CLI
        :param update_notes: documentation about the update
        :param koji_builds: list of koji builds or None (and pick latest)
        """
        logger.debug(
            f"Create bodhi update, "
            f"builds={koji_builds}, dg_branch={dist_git_branch}, type={update_type}"
        )
        self.dg.create_bodhi_update(
            koji_builds=koji_builds,
            dist_git_branch=dist_git_branch,
            update_notes=update_notes,
            update_type=update_type,
        )

    def create_srpm(
        self,
        output_file: str = None,
        upstream_ref: str = None,
        srpm_dir: Union[Path, str] = None,
    ) -> Path:
        """
        Create srpm from the upstream repo

        :param upstream_ref: git ref to upstream commit
        :param output_file: path + filename where the srpm should be written, defaults to cwd
        :param srpm_dir: path to the directory where the srpm is meant to be placed
        :return: a path to the srpm
        """
        self.up.run_action(actions=ActionName.post_upstream_clone)

        try:
            self.up.prepare_upstream_for_srpm_creation(upstream_ref=upstream_ref)
        except Exception as ex:
            raise PackitSRPMException(
                f"Preparation of the repository for creation of an SRPM failed: {ex}"
            ) from ex
        try:
            srpm_path = self.up.create_srpm(srpm_path=output_file, srpm_dir=srpm_dir)
        except PackitSRPMException:
            raise
        except Exception as ex:
            raise PackitSRPMException(
                f"An unexpected error occurred when creating the SRPM: {ex}"
            ) from ex

        if not srpm_path.exists():
            raise PackitSRPMNotFoundException(
                f"SRPM was created successfully, but can't be found at {srpm_path}"
            )
        return srpm_path

    def create_rpms(self, upstream_ref: str = None, rpm_dir: str = None) -> List[Path]:
        """
        Create rpms from the upstream repo

        :param upstream_ref: git ref to upstream commit
        :param rpm_dir: path to the directory where the rpm is meant to be placed
        :return: a path to the rpm
        """
        self.up.run_action(actions=ActionName.post_upstream_clone)

        try:
            self.up.prepare_upstream_for_srpm_creation(upstream_ref=upstream_ref)
        except Exception as ex:
            raise PackitRPMException(
                f"Preparing of the upstream to the RPM build failed: {ex}"
            ) from ex

        try:
            rpm_paths = self.up.create_rpms(rpm_dir=rpm_dir)
        except PackitRPMException:
            raise
        except Exception as ex:
            raise PackitRPMException(
                f"An unexpected error occurred when creating the RPMs: {ex}"
            ) from ex

        for rpm_path in rpm_paths:
            if not rpm_path.exists():
                raise PackitRPMNotFoundException(
                    f"RPM was created successfully, but can't be found at {rpm_path}"
                )
        return rpm_paths

    @staticmethod
    async def status_get_downstream_prs(status) -> List[Tuple[int, str, str]]:
        try:
            await asyncio.sleep(0)
            return status.get_downstream_prs()
        except Exception as exc:
            # https://github.com/packit/ogr/issues/67 work-around
            logger.error(f"Failed when getting downstream PRs: {exc}")
            return []

    @staticmethod
    async def status_get_dg_versions(status) -> Dict:
        try:
            await asyncio.sleep(0)
            return status.get_dg_versions()
        except Exception as exc:
            logger.error(f"Failed when getting Dist-git versions: {exc}")
            return {}

    @staticmethod
    async def status_get_up_releases(status) -> List:
        try:
            await asyncio.sleep(0)
            return status.get_up_releases()
        except Exception as exc:
            logger.error(f"Failed when getting upstream releases: {exc}")
            return []

    @staticmethod
    async def status_get_koji_builds(status) -> Dict:
        try:
            await asyncio.sleep(0)
            return status.get_koji_builds()
        except Exception as exc:
            logger.error(f"Failed when getting Koji builds: {exc}")
            return {}

    @staticmethod
    async def status_get_copr_builds(status) -> List:
        try:
            await asyncio.sleep(0)
            return status.get_copr_builds()
        except Exception as exc:
            logger.error(f"Failed when getting Copr builds: {exc}")
            return []

    @staticmethod
    async def status_get_updates(status) -> List:
        try:
            await asyncio.sleep(0)
            return status.get_updates()
        except Exception as exc:
            logger.error(f"Failed when getting Bodhi updates: {exc}")
            return []

    @staticmethod
    async def status_main(status: Status) -> List:
        """
        Schedule repository data retrieval calls concurrently.
        :param status: status of the package
        :return: awaitable tasks
        """
        res = await asyncio.gather(
            PackitAPI.status_get_downstream_prs(status),
            PackitAPI.status_get_dg_versions(status),
            PackitAPI.status_get_up_releases(status),
            PackitAPI.status_get_koji_builds(status),
            PackitAPI.status_get_copr_builds(status),
            PackitAPI.status_get_updates(status),
        )
        return res

    def status(self) -> None:
        status = Status(self.config, self.package_config, self.up, self.dg)
        if sys.version_info >= (3, 7, 0):
            res = asyncio.run(self.status_main(status))
        else:
            # backward compatibility for Python 3.6
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try:
                res = loop.run_until_complete(
                    asyncio.gather(
                        self.status_get_downstream_prs(status),
                        self.status_get_dg_versions(status),
                        self.status_get_up_releases(status),
                        self.status_get_koji_builds(status),
                        self.status_get_copr_builds(status),
                        self.status_get_updates(status),
                    )
                )
            finally:
                asyncio.set_event_loop(None)
                loop.close()

        (ds_prs, dg_versions, up_releases, koji_builds, copr_builds, updates) = res

        if ds_prs:
            logger.info("\nDownstream PRs:")
            logger.info(tabulate(ds_prs, headers=["ID", "Title", "URL"]))
        else:
            logger.info("\nNo downstream PRs found.")

        if dg_versions:
            logger.info("\nDist-git versions:")
            for branch, dg_version in dg_versions.items():
                logger.info(f"{branch}: {dg_version}")
        else:
            logger.info("\nNo Dist-git versions found.")

        if up_releases:
            logger.info("\nUpstream releases:")
            upstream_releases_str = "\n".join(
                f"{release.tag_name}" for release in up_releases
            )
            logger.info(upstream_releases_str)
        else:
            logger.info("\nNo upstream releases found.")

        if updates:
            logger.info("\nLatest Bodhi updates:")
            logger.info(tabulate(updates, headers=["Update", "Karma", "status"]))
        else:
            logger.info("\nNo Bodhi updates found.")

        if koji_builds:
            logger.info("\nLatest Koji builds:")
            for branch, branch_builds in koji_builds.items():
                logger.info(f"{branch}: {branch_builds}")
        else:
            logger.info("\nNo Koji builds found.")

        if copr_builds:
            logger.info("\nLatest Copr builds:")
            logger.info(
                tabulate(copr_builds, headers=["Build ID", "Project name", "Status"])
            )
        else:
            logger.info("\nNo Copr builds found.")

    def run_copr_build(
        self,
        project: str,
        chroots: List[str],
        owner: str = None,
        description: str = None,
        instructions: str = None,
        upstream_ref: str = None,
        list_on_homepage: bool = False,
        preserve_project: bool = False,
        additional_packages: List[str] = None,
        additional_repos: List[str] = None,
        request_admin_if_needed: bool = False,
    ) -> Tuple[int, str]:
        """
        Submit a build to copr build system using an SRPM using the current checkout.

        :param project: name of the copr project to build
                        inside (defaults to something long and ugly)
        :param chroots: a list of COPR chroots (targets) e.g. fedora-rawhide-x86_64
        :param owner: defaults to username from copr config file
        :param description: description of the project
        :param instructions: installation instructions for the project
        :param upstream_ref: git ref to upstream commit
        :param list_on_homepage: if set, created copr project will be visible on copr's home-page
        :param preserve_project: if set, project will not be created as temporary
        :param list additional_packages: buildroot packages for the chroot [DOES NOT WORK YET]
        :param list additional_repos: buildroot additional additional_repos
        :param bool request_admin_if_needed: if we can't change the settings
                                             and are not allowed to do so
        :return: id of the created build and url to the build web page
        """
        srpm_path = self.create_srpm(
            upstream_ref=upstream_ref, srpm_dir=self.up.local_project.working_dir
        )

        owner = owner or self.copr_helper.configured_owner
        if not owner:
            raise PackitCoprException(
                "Copr owner not set. Use Copr config file or `--owner` when calling packit CLI."
            )
        logger.info(f"We will operate with COPR owner {owner}.")

        self.copr_helper.create_copr_project_if_not_exists(
            project=project,
            chroots=chroots,
            owner=owner,
            description=description,
            instructions=instructions,
            list_on_homepage=list_on_homepage,
            preserve_project=preserve_project,
            additional_packages=additional_packages,
            additional_repos=additional_repos,
            request_admin_if_needed=request_admin_if_needed,
        )
        logger.debug(
            f"Submitting a build to copr build system,"
            f"owner={owner}, project={project}, path={srpm_path}"
        )

        build = self.copr_helper.copr_client.build_proxy.create_from_file(
            ownername=owner, projectname=project, path=srpm_path
        )
        return build.id, self.copr_helper.copr_web_build_url(build)

    def watch_copr_build(
        self, build_id: int, timeout: int, report_func: Callable = None
    ) -> str:
        """returns copr build state"""
        return self.copr_helper.watch_copr_build(
            build_id=build_id, timeout=timeout, report_func=report_func
        )

    @staticmethod
    def push_bodhi_update(update_alias: str):
        from bodhi.client.bindings import BodhiClient, UpdateNotFound

        b = BodhiClient()
        try:
            response = b.request(update=update_alias, request="stable")
            logger.debug(f"Bodhi response:\n{response}")
            response = response["update"]
            logger.info(
                f"Bodhi update {response['alias']} ({response['title']}) pushed to stable:\n"
                f"- {response['url']}\n"
                f"- karma: {response['karma']}\n"
                f"- notes:\n{response['notes']}\n"
            )
        except UpdateNotFound:
            logger.error("Update was not found.")

    def get_testing_updates(self, update_alias: Optional[str]) -> List:
        from bodhi.client.bindings import BodhiClient

        b = BodhiClient()
        results = b.query(
            alias=update_alias,
            packages=self.dg.package_config.downstream_package_name,
            status="testing",
        )["updates"]
        logger.debug("Bodhi updates with status 'testing' fetched.")

        return results

    @staticmethod
    def days_in_testing(update) -> int:
        if update.get("date_testing"):
            date_testing = datetime.strptime(
                update["date_testing"], "%Y-%m-%d %H:%M:%S"
            )
            return (datetime.utcnow() - date_testing).days
        return 0

    def push_updates(self, update_alias: Optional[str] = None):
        updates = self.get_testing_updates(update_alias)
        if not updates:
            logger.info("No testing updates found.")
        for update in updates:
            if self.days_in_testing(update) >= update["stable_days"]:
                self.push_bodhi_update(update["alias"])
            else:
                logger.debug(f"{update['alias']} is not ready to be pushed to stable")

    def init_kerberos_ticket(self) -> None:
        """
        Initialize the kerberos ticket if we have fas_user and keytab_path configured.

        The `kinit` command is run only once when called multiple times.
        """
        if self._kerberos_initialized:
            return
        self._run_kinit()
        self._kerberos_initialized = True

    def _run_kinit(self) -> None:
        """
        Run `kinit` if we have fas_user and keytab_path configured.
        """
        if (
            not self.config.fas_user
            or not self.config.keytab_path
            or not Path(self.config.keytab_path).is_file()
        ):
            logger.info("Won't be doing kinit, no credentials provided.")
            return

        cmd = [
            "kinit",
            f"{self.config.fas_user}@{self.config.kerberos_realm}",
            "-k",
            "-t",
            self.config.keytab_path,
        ]

        commands.run_command_remote(
            cmd=cmd,
            error_message="Failed to init kerberos ticket:",
            fail=True,
            # this prints debug logs from kerberos to stdout
            env={"KRB5_TRACE": "/dev/stdout"},
        )

    def clean(self):
        """clean up stuff once all the work is done"""
        # this is called in p-s: Handler.clean
        if self.up.is_command_handler_set():
            self.up.command_handler.clean()
        if self.dg.is_command_handler_set():
            self.dg.command_handler.clean()

    @staticmethod
    def validate_package_config(working_dir: Path) -> str:
        """validate .packit.yaml on the provided path and return human readable report"""
        config_path = find_packit_yaml(
            working_dir,
            try_local_dir_last=True,
        )
        config_content = load_packit_yaml(config_path)
        v = PackageConfigValidator(config_path, config_content)
        return v.validate()

    def create_sourcegit_from_upstream(
        self,
        upstream_url: Optional[str] = None,
        upstream_ref: Optional[str] = None,
        dist_git_path: Optional[Path] = None,
        dist_git_branch: Optional[str] = None,
        fedora_package: Optional[str] = None,
        centos_package: Optional[str] = None,
    ):
        """
        generate a source-git repo from provided upstream repo
        and a corresponding package in Fedora/CentOS ecosystem

        :param upstream_url: upstream repo URL we want to use as a base
        :param upstream_ref: upstream git-ref to use as a base
        :param fedora_package: pick up specfile and downstream sources from this fedora package
        :param centos_package: pick up specfile and downstream sources from this centos package
        :param dist_git_branch: branch in dist-git to use
        :param dist_git_path: path to a local clone of a dist-git repo
        """
        sgg = SourceGitGenerator(
            config=self.config,
            local_project=self.upstream_local_project,
            upstream_url=upstream_url,
            upstream_ref=upstream_ref,
            dist_git_path=dist_git_path,
            dist_git_branch=dist_git_branch,
            fedora_package=fedora_package,
            centos_package=centos_package,
        )
        sgg.create_from_upstream()



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest

from packit.actions import ActionName


@pytest.mark.parametrize(
    "action,valid",
    [
        ("get-current-version", True),
        ("create-patches", True),
        ("unknown-action", False),
        ("create_patches", False),
    ],
)
def test_is_valid(action, valid):
    assert ActionName.is_valid_action(action) == valid


def test_get_possible_values():
    values = ActionName.get_possible_values()
    assert values
    assert isinstance(values, list)
    for action_value in values:
        assert isinstance(action_value, str)
        assert "_" not in action_value


@pytest.mark.parametrize(
    "action_name,result",
    [
        ("get-current-version", ActionName.get_current_version),
        ("create-patches", ActionName.create_patches),
        ("unknown-action", None),
        ("create_patches", None),
    ],
)
def test_get_action_from_name(action_name, result):
    assert ActionName.get_action_from_name(action_name) == result



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from munch import Munch

BODHI_UPDATES = Munch(
    {
        "updates": [
            Munch(
                {
                    "title": "colin-0.3.1-1.fc29",
                    "autokarma": True,
                    "stable_karma": 3,
                    "unstable_karma": -3,
                    "requirements": "",
                    "require_bugs": True,
                    "require_testcases": True,
                    "display_name": "",
                    "notes": "new upstream release\n\n----\n\nlatest upstream "
                    "release\n\n----\n\nUpdate to latest upstream release.",
                    "type": "enhancement",
                    "status": "stable",
                    "request": None,
                    "severity": "unspecified",
                    "suggest": "unspecified",
                    "locked": False,
                    "pushed": True,
                    "critpath": False,
                    "close_bugs": True,
                    "date_submitted": "2019-01-21 16:51:00",
                    "date_modified": None,
                    "date_approved": None,
                    "date_pushed": "2019-01-30 02:06:12",
                    "date_testing": "2019-01-22 03:03:10",
                    "date_stable": "2019-01-30 02:06:12",
                    "alias": "FEDORA-2019-7006fbed73",
                    "old_updateid": None,
                    "test_gating_status": "ignored",
                    "greenwave_summary_string": "no tests are required",
                    "greenwave_unsatisfied_requirements": None,
                    "meets_testing_requirements": True,
                    "url": "https://bodhi.fedoraproject.org/updates/FEDORA-2019-7006fbed73",
                    "release": Munch(
                        {
                            "name": "F29",
                            "long_name": "Fedora 29",
                            "version": "29",
                            "id_prefix": "FEDORA",
                            "branch": "f29",
                            "dist_tag": "f29",
                            "stable_tag": "f29-updates",
                            "testing_tag": "f29-updates-testing",
                            "candidate_tag": "f29-updates-candidate",
                            "pending_signing_tag": "f29-signing-pending",
                            "pending_testing_tag": "f29-updates-testing-pending",
                            "pending_stable_tag": "f29-updates-pending",
                            "override_tag": "f29-override",
                            "mail_template": "fedora_errata_template",
                            "state": "current",
                            "composed_by_bodhi": True,
                            "composes": [],
                        }
                    ),
                    "comments": [
                        Munch(
                            {
                                "id": 886725,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been submitted "
                                "for testing by ttomecek. ",
                                "anonymous": False,
                                "timestamp": "2019-01-21 16:51:00",
                                "update_id": 130273,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 886729,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has obsoleted [colin-0.3.0-2.fc29]"
                                "(https://bodhi.fedoraproject.org/updates/"
                                "FEDORA-2019-f921969692), and has inherited "
                                "its bugs and notes.",
                                "anonymous": False,
                                "timestamp": "2019-01-21 16:51:03",
                                "update_id": 130273,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 886745,
                                "karma": 1,
                                "karma_critpath": 0,
                                "text": "Works well, thanks!\n\n(tested with "
                                "registry.fedoraproject.org/fedora:29 image)",
                                "anonymous": False,
                                "timestamp": "2019-01-21 18:09:32",
                                "update_id": 130273,
                                "user_id": 4222,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 4222,
                                        "name": "lachmanfrantisek",
                                        "email": "flachman@redhat.com",
                                        "show_popups": True,
                                        "avatar": "https://seccdn.libravatar.org/avatar/"
                                        "38deedcf02e4617a9a8a0c04b67fd4c7fcea9"
                                        "ededf02d13a93a7626477350e45?s=24&d=retro",
                                        "openid": "lachmanfrantisek.id.fedoraproject.org",
                                        "groups": [Munch({"name": "packager"})],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 886940,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been pushed to testing.",
                                "anonymous": False,
                                "timestamp": "2019-01-22 03:04:32",
                                "update_id": 130273,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 889233,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has reached 7 days in testing "
                                "and can be pushed to stable now "
                                "if the maintainer wishes",
                                "anonymous": False,
                                "timestamp": "2019-01-29 06:00:45",
                                "update_id": 130273,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 889253,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been submitted "
                                "for batched by ttomecek.",
                                "anonymous": False,
                                "timestamp": "2019-01-29 08:41:38",
                                "update_id": 130273,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 889436,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been submitted for stable by bodhi. ",
                                "anonymous": False,
                                "timestamp": "2019-01-29 23:45:40",
                                "update_id": 130273,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 889521,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been pushed to stable.",
                                "anonymous": False,
                                "timestamp": "2019-01-30 02:07:18",
                                "update_id": 130273,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                    ],
                    "builds": [
                        Munch(
                            {
                                "nvr": "colin-0.3.1-1.fc29",
                                "release_id": 23,
                                "signed": True,
                                "ci_url": None,
                                "type": "rpm",
                                "epoch": 0,
                            }
                        )
                    ],
                    "compose": None,
                    "bugs": [],
                    "user": Munch(
                        {
                            "id": 754,
                            "name": "ttomecek",
                            "email": "ttomecek@redhat.com",
                            "show_popups": True,
                            "avatar": "https://seccdn.libravatar.org/avatar/"
                            "2098fd3a827c38e855bd76b1f6b75c85483b1"
                            "12f0e8f086257309df75dc04dc0?s=24&d=retro",
                            "openid": "ttomecek.id.fedoraproject.org",
                            "groups": [
                                Munch({"name": "provenpackager"}),
                                Munch({"name": "packager"}),
                            ],
                        }
                    ),
                    "updateid": "FEDORA-2019-7006fbed73",
                    "submitter": "ttomecek",
                    "karma": 1,
                    "content_type": "rpm",
                    "test_cases": [],
                }
            ),
            Munch(
                {
                    "title": "colin-0.3.1-1.fc28",
                    "autokarma": True,
                    "stable_karma": 3,
                    "unstable_karma": -3,
                    "requirements": "",
                    "require_bugs": True,
                    "require_testcases": True,
                    "display_name": "",
                    "notes": "new upstream release\n\n----\n\nlatest upstream "
                    "release\n\n----\n\nUpdate to latest upstream release.",
                    "type": "enhancement",
                    "status": "stable",
                    "request": None,
                    "severity": "unspecified",
                    "suggest": "unspecified",
                    "locked": False,
                    "pushed": True,
                    "critpath": False,
                    "close_bugs": True,
                    "date_submitted": "2019-01-21 16:50:59",
                    "date_modified": None,
                    "date_approved": None,
                    "date_pushed": "2019-01-30 01:31:27",
                    "date_testing": "2019-01-22 01:16:27",
                    "date_stable": "2019-01-30 01:31:27",
                    "alias": "FEDORA-2019-cb1e057344",
                    "old_updateid": None,
                    "test_gating_status": "ignored",
                    "greenwave_summary_string": "no tests are required",
                    "greenwave_unsatisfied_requirements": None,
                    "meets_testing_requirements": True,
                    "url": "https://bodhi.fedoraproject.org/updates/FEDORA-2019-cb1e057344",
                    "release": Munch(
                        {
                            "name": "F28",
                            "long_name": "Fedora 28",
                            "version": "28",
                            "id_prefix": "FEDORA",
                            "branch": "f28",
                            "dist_tag": "f28",
                            "stable_tag": "f28-updates",
                            "testing_tag": "f28-updates-testing",
                            "candidate_tag": "f28-updates-candidate",
                            "pending_signing_tag": "f28-signing-pending",
                            "pending_testing_tag": "f28-updates-testing-pending",
                            "pending_stable_tag": "f28-updates-pending",
                            "override_tag": "f28-override",
                            "mail_template": "fedora_errata_template",
                            "state": "current",
                            "composed_by_bodhi": True,
                            "composes": [],
                        }
                    ),
                    "comments": [
                        Munch(
                            {
                                "id": 886724,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been submitted "
                                "for testing by ttomecek. ",
                                "anonymous": False,
                                "timestamp": "2019-01-21 16:50:59",
                                "update_id": 130272,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 886727,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has obsoleted [colin-0.3.0-2.fc28]"
                                "(https://bodhi.fedoraproject.org/updates/"
                                "FEDORA-2019-7223843531), and has inherited "
                                "its bugs and notes.",
                                "anonymous": False,
                                "timestamp": "2019-01-21 16:51:02",
                                "update_id": 130272,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 886744,
                                "karma": 1,
                                "karma_critpath": 0,
                                "text": "Works well, thanks!\n\n(tested with "
                                "registry.fedoraproject.org/fedora:28 image)",
                                "anonymous": False,
                                "timestamp": "2019-01-21 18:08:52",
                                "update_id": 130272,
                                "user_id": 4222,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 4222,
                                        "name": "lachmanfrantisek",
                                        "email": "flachman@redhat.com",
                                        "show_popups": True,
                                        "avatar": "https://seccdn.libravatar.org/avatar/"
                                        "38deedcf02e4617a9a8a0c04b67fd4c7fcea9"
                                        "ededf02d13a93a7626477350e45?s=24&d=retro",
                                        "openid": "lachmanfrantisek.id.fedoraproject.org",
                                        "groups": [Munch({"name": "packager"})],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 886870,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been pushed to testing.",
                                "anonymous": False,
                                "timestamp": "2019-01-22 01:17:03",
                                "update_id": 130272,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 889232,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has reached 7 days in testing "
                                "and can be pushed to stable now "
                                "if the maintainer wishes",
                                "anonymous": False,
                                "timestamp": "2019-01-29 06:00:44",
                                "update_id": 130272,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 889254,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been submitted for "
                                "batched by ttomecek. ",
                                "anonymous": False,
                                "timestamp": "2019-01-29 08:41:41",
                                "update_id": 130272,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 889435,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been submitted for stable by bodhi. ",
                                "anonymous": False,
                                "timestamp": "2019-01-29 23:45:39",
                                "update_id": 130272,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 889457,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been pushed to stable.",
                                "anonymous": False,
                                "timestamp": "2019-01-30 01:32:23",
                                "update_id": 130272,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                    ],
                    "builds": [
                        Munch(
                            {
                                "nvr": "colin-0.3.1-1.fc28",
                                "release_id": 21,
                                "signed": True,
                                "ci_url": None,
                                "type": "rpm",
                                "epoch": 0,
                            }
                        )
                    ],
                    "compose": None,
                    "bugs": [],
                    "user": Munch(
                        {
                            "id": 754,
                            "name": "ttomecek",
                            "email": "ttomecek@redhat.com",
                            "show_popups": True,
                            "avatar": "https://seccdn.libravatar.org/avatar/"
                            "2098fd3a827c38e855bd76b1f6b75c85483b1"
                            "12f0e8f086257309df75dc04dc0?s=24&d=retro",
                            "openid": "ttomecek.id.fedoraproject.org",
                            "groups": [
                                Munch({"name": "provenpackager"}),
                                Munch({"name": "packager"}),
                            ],
                        }
                    ),
                    "updateid": "FEDORA-2019-cb1e057344",
                    "submitter": "ttomecek",
                    "karma": 1,
                    "content_type": "rpm",
                    "test_cases": [],
                }
            ),
            Munch(
                {
                    "title": "colin-0.3.0-2.fc28",
                    "autokarma": True,
                    "stable_karma": 3,
                    "unstable_karma": -3,
                    "requirements": "",
                    "require_bugs": True,
                    "require_testcases": True,
                    "display_name": "",
                    "notes": "latest upstream release\n\n----\n\nUpdate to "
                    "latest upstream release.",
                    "type": "enhancement",
                    "status": "obsolete",
                    "request": None,
                    "severity": "unspecified",
                    "suggest": "unspecified",
                    "locked": False,
                    "pushed": False,
                    "critpath": False,
                    "close_bugs": True,
                    "date_submitted": "2019-01-18 10:38:22",
                    "date_modified": None,
                    "date_approved": None,
                    "date_pushed": "2019-01-19 01:43:21",
                    "date_testing": "2019-01-19 01:43:21",
                    "date_stable": None,
                    "alias": "FEDORA-2019-7223843531",
                    "old_updateid": None,
                    "test_gating_status": "ignored",
                    "greenwave_summary_string": "no tests are required",
                    "greenwave_unsatisfied_requirements": None,
                    "meets_testing_requirements": True,
                    "url": "https://bodhi.fedoraproject.org/updates/FEDORA-2019-7223843531",
                    "release": Munch(
                        {
                            "name": "F28",
                            "long_name": "Fedora 28",
                            "version": "28",
                            "id_prefix": "FEDORA",
                            "branch": "f28",
                            "dist_tag": "f28",
                            "stable_tag": "f28-updates",
                            "testing_tag": "f28-updates-testing",
                            "candidate_tag": "f28-updates-candidate",
                            "pending_signing_tag": "f28-signing-pending",
                            "pending_testing_tag": "f28-updates-testing-pending",
                            "pending_stable_tag": "f28-updates-pending",
                            "override_tag": "f28-override",
                            "mail_template": "fedora_errata_template",
                            "state": "current",
                            "composed_by_bodhi": True,
                            "composes": [],
                        }
                    ),
                    "comments": [
                        Munch(
                            {
                                "id": 885830,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been submitted "
                                "for testing by ttomecek. ",
                                "anonymous": False,
                                "timestamp": "2019-01-18 10:38:22",
                                "update_id": 130128,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 885834,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has obsoleted [colin-0.2.1-1.fc28]"
                                "(https://bodhi.fedoraproject.org/updates/"
                                "FEDORA-2018-507426b4b6), "
                                "and has inherited its bugs and notes.",
                                "anonymous": False,
                                "timestamp": "2019-01-18 10:38:25",
                                "update_id": 130128,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 886062,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been pushed to testing.",
                                "anonymous": False,
                                "timestamp": "2019-01-19 01:44:06",
                                "update_id": 130128,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/img/"
                                        "icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                        Munch(
                            {
                                "id": 886726,
                                "karma": 0,
                                "karma_critpath": 0,
                                "text": "This update has been obsoleted by [colin-0.3.1-1.fc28]"
                                "(https://bodhi.fedoraproject.org/updates/"
                                "FEDORA-2019-cb1e057344).",
                                "anonymous": False,
                                "timestamp": "2019-01-21 16:51:02",
                                "update_id": 130128,
                                "user_id": 91,
                                "bug_feedback": [],
                                "testcase_feedback": [],
                                "user": Munch(
                                    {
                                        "id": 91,
                                        "name": "bodhi",
                                        "email": None,
                                        "show_popups": True,
                                        "avatar": "https://apps.fedoraproject.org/"
                                        "img/icons/bodhi-24.png",
                                        "openid": "bodhi.id.fedoraproject.org",
                                        "groups": [],
                                    }
                                ),
                            }
                        ),
                    ],
                    "builds": [
                        Munch(
                            {
                                "nvr": "colin-0.3.0-2.fc28",
                                "release_id": 21,
                                "signed": True,
                                "ci_url": None,
                                "type": "rpm",
                                "epoch": 0,
                            }
                        )
                    ],
                    "compose": None,
                    "bugs": [],
                    "user": Munch(
                        {
                            "id": 754,
                            "name": "ttomecek",
                            "email": "ttomecek@redhat.com",
                            "show_popups": True,
                            "avatar": "https://seccdn.libravatar.org/avatar/"
                            "2098fd3a827c38e855bd76b1f6b75c85483b112"
                            "f0e8f086257309df75dc04dc0?s=24&d=retro",
                            "openid": "ttomecek.id.fedoraproject.org",
                            "groups": [
                                Munch({"name": "provenpackager"}),
                                Munch({"name": "packager"}),
                            ],
                        }
                    ),
                    "updateid": "FEDORA-2019-7223843531",
                    "submitter": "ttomecek",
                    "karma": 0,
                    "content_type": "rpm",
                    "test_cases": [],
                }
            ),
        ],
        "page": 1,
        "pages": 1,
        "rows_per_page": 20,
        "total": 11,
        "chrome": True,
        "display_user": True,
        "display_request": True,
        "package": "colin",
        "active_releases": False,
    }
)



# MIT License
#
# Copyright (c) 2019 Red Hat, Inc.

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pytest
from pkg_resources import get_distribution

from packit.cli.build import build
from packit.cli.create_update import create_update
from packit.cli.packit_base import packit_base
from packit.cli.propose_downstream import propose_downstream
from tests.spellbook import call_packit


def test_base_help():
    result = call_packit(parameters=["--help"])
    assert result.exit_code == 0
    assert "Usage: packit [OPTIONS] COMMAND [ARGS]..." in result.output


def test_base_version():
    # This test requires packit on pythonpath
    result = call_packit(parameters=["--version"])
    assert result.exit_code == 0
    assert result.output.strip() == get_distribution("packitos").version


@pytest.mark.parametrize("cmd_function", [propose_downstream, build, create_update])
def test_base_subcommand_direct(cmd_function):
    result = call_packit(cmd_function, parameters=["--help"])
    assert result.exit_code == 0


@pytest.mark.parametrize("subcommand", ["propose-downstream", "build", "create-update"])
def test_base_subcommand_help(subcommand):
    result = call_packit(packit_base, parameters=[subcommand, "--help"])
    assert result.exit_code == 0
    assert f"Usage: packit {subcommand} [OPTIONS]" in result.output
